/*
 *  si-82-sectrust-ct.c
 *  Security
 *
 *  Copyright (c) 2014 Apple Inc. All Rights Reserved.
 *
 */

#include <CoreFoundation/CoreFoundation.h>
#include <Security/SecCertificatePriv.h>
#include <Security/SecTrustPriv.h>
#include <Security/SecPolicy.h>
#include <stdlib.h>
#include <unistd.h>
#include <utilities/SecCFRelease.h>
#include <utilities/SecCFWrappers.h>

#include "shared_regressions.h"

#include "si-82-sectrust-ct-certs.h"


static void test_ct_trust(CFArrayRef certs, CFArrayRef scts, CFTypeRef ocspresponses, CFArrayRef anchors, CFArrayRef trustedLogs, CFStringRef hostname, bool expected, const char *test_name)
{
    CFDateRef date=NULL;
    CFArrayRef policies=NULL;
    SecPolicyRef policy=NULL;
    SecTrustRef trust=NULL;
    SecTrustResultType trustResult;
    CFCalendarRef cal = NULL;
    CFAbsoluteTime at;
    CFDictionaryRef results;
    CFArrayRef properties;



    isnt(policy = SecPolicyCreateSSL(false, hostname), NULL, "create policy");
    isnt(policies = CFArrayCreate(kCFAllocatorDefault, (const void **)&policy, 1, &kCFTypeArrayCallBacks), NULL, "create policies");
    ok_status(SecTrustCreateWithCertificates(certs, policies, &trust), "create trust");

    isnt(cal = CFCalendarCreateWithIdentifier(kCFAllocatorDefault, kCFGregorianCalendar), NULL, "create calendar");
    ok(CFCalendarComposeAbsoluteTime(cal, &at, "yMd", 2015, 3, 7), "create verify absolute time");
    isnt(date = CFDateCreate(kCFAllocatorDefault, at), NULL, "create verify date");

    ok_status(SecTrustSetAnchorCertificates(trust, anchors), "set anchors");

    ok_status(SecTrustSetSignedCertificateTimestamps(trust, scts), "set standalone SCTs");;

    if(trustedLogs) {
        ok_status(SecTrustSetTrustedLogs(trust, trustedLogs), "set trusted logs");
    }

    if(ocspresponses) {
        ok_status(SecTrustSetOCSPResponse(trust, ocspresponses), "set ocsp responses");
    }

    ok_status(SecTrustSetVerifyDate(trust, date), "set date");
    ok_status(SecTrustEvaluate(trust, &trustResult), "evaluate trust");
    ok(trustResult == kSecTrustResultUnspecified, "trustResult 4 expected (got %d)",
       (int)trustResult);

    results = SecTrustCopyResult(trust);

    CFTypeRef ct = CFDictionaryGetValue(results, kSecTrustCertificateTransparency);
    bool rv  = false;
    if (ct && CFGetTypeID(ct) == CFBooleanGetTypeID()) {
        rv = CFBooleanGetValue((CFBooleanRef)ct);
    }
    ok((rv == expected), "unexpected CT result (%s)", test_name);

#if 0
    // For test debugging purposes:
    if(ct) {
        printf("CT result:\n");
        CFShow(ct);
    } else {
        printf("no CT result\n");
    }
#endif

#if 0
    // for later: we will want to check that EV result abide by CT rules

    CFTypeRef ev = CFDictionaryGetValue(results, kSecTrustExtendedValidation);

    if(ev) {
        printf("EV result:\n");
        CFShow(ev);
    } else {
        printf("no EV result\n");
    }
#endif

    properties = SecTrustCopyProperties(trust);

    CFReleaseSafe(policy);
    CFReleaseSafe(policies);
    CFReleaseSafe(trust);
    CFReleaseSafe(date);
    CFReleaseSafe(cal);
    CFReleaseSafe(results);
    CFReleaseSafe(properties);
}

static
unsigned char CA_alpha_cert_der[] = {
    0x30, 0x82, 0x02, 0xc7, 0x30, 0x82, 0x02, 0x30, 0xa0, 0x03, 0x02, 0x01,
    0x02, 0x02, 0x01, 0x01, 0x30, 0x0d, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86,
    0xf7, 0x0d, 0x01, 0x01, 0x05, 0x05, 0x00, 0x30, 0x52, 0x31, 0x0b, 0x30,
    0x09, 0x06, 0x03, 0x55, 0x04, 0x06, 0x13, 0x02, 0x55, 0x53, 0x31, 0x1a,
    0x30, 0x18, 0x06, 0x03, 0x55, 0x04, 0x0a, 0x0c, 0x11, 0x63, 0x6f, 0x72,
    0x65, 0x6f, 0x73, 0x2d, 0x63, 0x74, 0x2d, 0x74, 0x65, 0x73, 0x74, 0x20,
    0x43, 0x41, 0x31, 0x13, 0x30, 0x11, 0x06, 0x03, 0x55, 0x04, 0x08, 0x0c,
    0x0a, 0x43, 0x61, 0x6c, 0x69, 0x66, 0x6f, 0x72, 0x6e, 0x69, 0x61, 0x31,
    0x12, 0x30, 0x10, 0x06, 0x03, 0x55, 0x04, 0x07, 0x0c, 0x09, 0x43, 0x75,
    0x70, 0x65, 0x72, 0x74, 0x69, 0x6e, 0x6f, 0x30, 0x1e, 0x17, 0x0d, 0x31,
    0x32, 0x30, 0x36, 0x30, 0x31, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x5a,
    0x17, 0x0d, 0x32, 0x32, 0x30, 0x36, 0x30, 0x31, 0x30, 0x30, 0x30, 0x30,
    0x30, 0x30, 0x5a, 0x30, 0x52, 0x31, 0x0b, 0x30, 0x09, 0x06, 0x03, 0x55,
    0x04, 0x06, 0x13, 0x02, 0x55, 0x53, 0x31, 0x1a, 0x30, 0x18, 0x06, 0x03,
    0x55, 0x04, 0x0a, 0x0c, 0x11, 0x63, 0x6f, 0x72, 0x65, 0x6f, 0x73, 0x2d,
    0x63, 0x74, 0x2d, 0x74, 0x65, 0x73, 0x74, 0x20, 0x43, 0x41, 0x31, 0x13,
    0x30, 0x11, 0x06, 0x03, 0x55, 0x04, 0x08, 0x0c, 0x0a, 0x43, 0x61, 0x6c,
    0x69, 0x66, 0x6f, 0x72, 0x6e, 0x69, 0x61, 0x31, 0x12, 0x30, 0x10, 0x06,
    0x03, 0x55, 0x04, 0x07, 0x0c, 0x09, 0x43, 0x75, 0x70, 0x65, 0x72, 0x74,
    0x69, 0x6e, 0x6f, 0x30, 0x81, 0x9f, 0x30, 0x0d, 0x06, 0x09, 0x2a, 0x86,
    0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01, 0x01, 0x05, 0x00, 0x03, 0x81, 0x8d,
    0x00, 0x30, 0x81, 0x89, 0x02, 0x81, 0x81, 0x00, 0xee, 0x58, 0x5c, 0xf8,
    0x95, 0x77, 0x15, 0x42, 0xcb, 0x3a, 0x42, 0x07, 0x31, 0x69, 0xea, 0xa6,
    0x7e, 0x73, 0x1c, 0x7c, 0x68, 0x2a, 0x07, 0xdc, 0xc2, 0x15, 0xed, 0xef,
    0x06, 0x3f, 0x94, 0x56, 0xa7, 0xce, 0x34, 0x59, 0xeb, 0x9f, 0xa8, 0xf1,
    0x12, 0x18, 0x57, 0xc2, 0xe5, 0xce, 0x69, 0x30, 0xbe, 0x6c, 0x45, 0x89,
    0x9b, 0x1a, 0x74, 0xbf, 0xe4, 0x33, 0xca, 0xf7, 0x1e, 0xb7, 0x7d, 0x94,
    0x90, 0x73, 0x51, 0xd4, 0x01, 0x22, 0x4f, 0x4e, 0x9d, 0x78, 0x1d, 0x7c,
    0x18, 0x3a, 0x99, 0x64, 0x9c, 0xf0, 0x10, 0x7b, 0xd2, 0xe9, 0x86, 0x1f,
    0x45, 0xc9, 0x86, 0x6c, 0x48, 0x5e, 0xab, 0x3d, 0xfb, 0xa6, 0xef, 0x45,
    0x5e, 0x23, 0x66, 0x8a, 0xd1, 0x61, 0x5d, 0x6c, 0x5e, 0x1d, 0xcf, 0xcc,
    0x54, 0xac, 0xf9, 0xca, 0xa8, 0xa7, 0x2d, 0xd1, 0xbf, 0xd8, 0xc7, 0xde,
    0x12, 0x68, 0x86, 0x5d, 0x02, 0x03, 0x01, 0x00, 0x01, 0xa3, 0x81, 0xac,
    0x30, 0x81, 0xa9, 0x30, 0x1d, 0x06, 0x03, 0x55, 0x1d, 0x0e, 0x04, 0x16,
    0x04, 0x14, 0xdc, 0x16, 0x44, 0x15, 0x3e, 0x53, 0x27, 0xd8, 0x68, 0x66,
    0x41, 0x40, 0x88, 0x90, 0xe4, 0x4e, 0x0a, 0xda, 0x08, 0xa9, 0x30, 0x7a,
    0x06, 0x03, 0x55, 0x1d, 0x23, 0x04, 0x73, 0x30, 0x71, 0x80, 0x14, 0xdc,
    0x16, 0x44, 0x15, 0x3e, 0x53, 0x27, 0xd8, 0x68, 0x66, 0x41, 0x40, 0x88,
    0x90, 0xe4, 0x4e, 0x0a, 0xda, 0x08, 0xa9, 0xa1, 0x56, 0xa4, 0x54, 0x30,
    0x52, 0x31, 0x0b, 0x30, 0x09, 0x06, 0x03, 0x55, 0x04, 0x06, 0x13, 0x02,
    0x55, 0x53, 0x31, 0x1a, 0x30, 0x18, 0x06, 0x03, 0x55, 0x04, 0x0a, 0x0c,
    0x11, 0x63, 0x6f, 0x72, 0x65, 0x6f, 0x73, 0x2d, 0x63, 0x74, 0x2d, 0x74,
    0x65, 0x73, 0x74, 0x20, 0x43, 0x41, 0x31, 0x13, 0x30, 0x11, 0x06, 0x03,
    0x55, 0x04, 0x08, 0x0c, 0x0a, 0x43, 0x61, 0x6c, 0x69, 0x66, 0x6f, 0x72,
    0x6e, 0x69, 0x61, 0x31, 0x12, 0x30, 0x10, 0x06, 0x03, 0x55, 0x04, 0x07,
    0x0c, 0x09, 0x43, 0x75, 0x70, 0x65, 0x72, 0x74, 0x69, 0x6e, 0x6f, 0x82,
    0x01, 0x01, 0x30, 0x0c, 0x06, 0x03, 0x55, 0x1d, 0x13, 0x04, 0x05, 0x30,
    0x03, 0x01, 0x01, 0xff, 0x30, 0x0d, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86,
    0xf7, 0x0d, 0x01, 0x01, 0x05, 0x05, 0x00, 0x03, 0x81, 0x81, 0x00, 0x96,
    0xeb, 0xd2, 0xcb, 0x18, 0x96, 0x4d, 0x63, 0xa5, 0xf4, 0x41, 0x18, 0xd6,
    0x52, 0x2a, 0xcf, 0xb2, 0x13, 0x5b, 0x44, 0x95, 0x17, 0xc3, 0x93, 0x4c,
    0x9b, 0x37, 0xdf, 0xa5, 0x8d, 0x9f, 0x34, 0x63, 0x93, 0xb2, 0x13, 0x28,
    0x0c, 0x17, 0xc6, 0xe9, 0x1d, 0xa9, 0xba, 0x4f, 0x7a, 0x58, 0x8d, 0x61,
    0xf5, 0xb4, 0x36, 0x25, 0xf9, 0x14, 0x38, 0x00, 0x53, 0x97, 0x98, 0x2e,
    0xd3, 0x56, 0xfd, 0x5d, 0x47, 0x97, 0x5c, 0xeb, 0xd8, 0x39, 0x2e, 0x77,
    0xd9, 0x44, 0x43, 0x8c, 0x11, 0x10, 0x93, 0x84, 0x41, 0x02, 0x5f, 0x85,
    0x28, 0xe7, 0xd3, 0x78, 0x76, 0x21, 0x82, 0x4c, 0xf5, 0xee, 0x87, 0x5d,
    0x9b, 0x78, 0x3a, 0x88, 0xeb, 0x65, 0xd6, 0x65, 0x76, 0x23, 0x32, 0xbf,
    0xaa, 0xfc, 0xe8, 0x9b, 0xad, 0x8d, 0xec, 0x22, 0x3b, 0x44, 0x4d, 0x8d,
    0xf8, 0x9b, 0x68, 0x2b, 0xa3, 0x53, 0xae
};


static
unsigned char CA_beta_cert_der[] = {
    0x30, 0x82, 0x02, 0xc7, 0x30, 0x82, 0x02, 0x30, 0xa0, 0x03, 0x02, 0x01,
    0x02, 0x02, 0x01, 0x01, 0x30, 0x0d, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86,
    0xf7, 0x0d, 0x01, 0x01, 0x05, 0x05, 0x00, 0x30, 0x52, 0x31, 0x0b, 0x30,
    0x09, 0x06, 0x03, 0x55, 0x04, 0x06, 0x13, 0x02, 0x55, 0x53, 0x31, 0x1a,
    0x30, 0x18, 0x06, 0x03, 0x55, 0x04, 0x0a, 0x0c, 0x11, 0x63, 0x6f, 0x72,
    0x65, 0x6f, 0x73, 0x2d, 0x63, 0x74, 0x2d, 0x74, 0x65, 0x73, 0x74, 0x20,
    0x43, 0x41, 0x31, 0x13, 0x30, 0x11, 0x06, 0x03, 0x55, 0x04, 0x08, 0x0c,
    0x0a, 0x43, 0x61, 0x6c, 0x69, 0x66, 0x6f, 0x72, 0x6e, 0x69, 0x61, 0x31,
    0x12, 0x30, 0x10, 0x06, 0x03, 0x55, 0x04, 0x07, 0x0c, 0x09, 0x43, 0x75,
    0x70, 0x65, 0x72, 0x74, 0x69, 0x6e, 0x6f, 0x30, 0x1e, 0x17, 0x0d, 0x31,
    0x35, 0x30, 0x33, 0x30, 0x35, 0x30, 0x30, 0x33, 0x34, 0x33, 0x33, 0x5a,
    0x17, 0x0d, 0x31, 0x36, 0x30, 0x33, 0x30, 0x34, 0x30, 0x30, 0x33, 0x34,
    0x33, 0x33, 0x5a, 0x30, 0x52, 0x31, 0x0b, 0x30, 0x09, 0x06, 0x03, 0x55,
    0x04, 0x06, 0x13, 0x02, 0x55, 0x53, 0x31, 0x1a, 0x30, 0x18, 0x06, 0x03,
    0x55, 0x04, 0x0a, 0x0c, 0x11, 0x63, 0x6f, 0x72, 0x65, 0x6f, 0x73, 0x2d,
    0x63, 0x74, 0x2d, 0x74, 0x65, 0x73, 0x74, 0x20, 0x43, 0x41, 0x31, 0x13,
    0x30, 0x11, 0x06, 0x03, 0x55, 0x04, 0x08, 0x0c, 0x0a, 0x43, 0x61, 0x6c,
    0x69, 0x66, 0x6f, 0x72, 0x6e, 0x69, 0x61, 0x31, 0x12, 0x30, 0x10, 0x06,
    0x03, 0x55, 0x04, 0x07, 0x0c, 0x09, 0x43, 0x75, 0x70, 0x65, 0x72, 0x74,
    0x69, 0x6e, 0x6f, 0x30, 0x81, 0x9f, 0x30, 0x0d, 0x06, 0x09, 0x2a, 0x86,
    0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01, 0x01, 0x05, 0x00, 0x03, 0x81, 0x8d,
    0x00, 0x30, 0x81, 0x89, 0x02, 0x81, 0x81, 0x00, 0xca, 0xab, 0x48, 0xae,
    0xd3, 0x7a, 0x27, 0x8b, 0x7c, 0x11, 0xb5, 0x73, 0xdb, 0x23, 0xba, 0xfc,
    0xb3, 0x7a, 0x49, 0x92, 0xd3, 0x2d, 0xbe, 0x31, 0x6b, 0x53, 0xd3, 0x78,
    0x8b, 0xf3, 0xc9, 0x77, 0x66, 0x53, 0xb1, 0xa2, 0xd8, 0xba, 0x85, 0xd3,
    0x6a, 0x2e, 0x9d, 0x68, 0xc1, 0x3b, 0x69, 0x6a, 0x2d, 0xf2, 0xc1, 0xc3,
    0xce, 0xcf, 0x38, 0x56, 0x92, 0x1a, 0x47, 0x9d, 0xdd, 0x59, 0x87, 0xb4,
    0x23, 0x8c, 0xbd, 0x0b, 0x21, 0x63, 0x19, 0x5e, 0x7d, 0x2d, 0x7a, 0x20,
    0xc4, 0x16, 0xc7, 0x29, 0x73, 0x0e, 0x43, 0x7b, 0xc1, 0xb7, 0xbb, 0xd9,
    0x8f, 0x24, 0x0a, 0xec, 0x52, 0x53, 0xef, 0xa2, 0xb2, 0x77, 0x8f, 0x38,
    0x52, 0x5e, 0x2f, 0xa0, 0xc7, 0x4d, 0x98, 0x66, 0xc1, 0xb0, 0x55, 0x03,
    0xb8, 0x6c, 0x32, 0x65, 0x67, 0xc4, 0xbd, 0xd9, 0x86, 0x83, 0x0f, 0x40,
    0x52, 0xde, 0xcd, 0x8b, 0x02, 0x03, 0x01, 0x00, 0x01, 0xa3, 0x81, 0xac,
    0x30, 0x81, 0xa9, 0x30, 0x1d, 0x06, 0x03, 0x55, 0x1d, 0x0e, 0x04, 0x16,
    0x04, 0x14, 0x64, 0x1f, 0x09, 0x99, 0x2d, 0x6a, 0x5b, 0x4d, 0xef, 0xed,
    0xbb, 0xba, 0x96, 0xf9, 0x73, 0x65, 0xad, 0x6e, 0x84, 0xbd, 0x30, 0x7a,
    0x06, 0x03, 0x55, 0x1d, 0x23, 0x04, 0x73, 0x30, 0x71, 0x80, 0x14, 0x64,
    0x1f, 0x09, 0x99, 0x2d, 0x6a, 0x5b, 0x4d, 0xef, 0xed, 0xbb, 0xba, 0x96,
    0xf9, 0x73, 0x65, 0xad, 0x6e, 0x84, 0xbd, 0xa1, 0x56, 0xa4, 0x54, 0x30,
    0x52, 0x31, 0x0b, 0x30, 0x09, 0x06, 0x03, 0x55, 0x04, 0x06, 0x13, 0x02,
    0x55, 0x53, 0x31, 0x1a, 0x30, 0x18, 0x06, 0x03, 0x55, 0x04, 0x0a, 0x0c,
    0x11, 0x63, 0x6f, 0x72, 0x65, 0x6f, 0x73, 0x2d, 0x63, 0x74, 0x2d, 0x74,
    0x65, 0x73, 0x74, 0x20, 0x43, 0x41, 0x31, 0x13, 0x30, 0x11, 0x06, 0x03,
    0x55, 0x04, 0x08, 0x0c, 0x0a, 0x43, 0x61, 0x6c, 0x69, 0x66, 0x6f, 0x72,
    0x6e, 0x69, 0x61, 0x31, 0x12, 0x30, 0x10, 0x06, 0x03, 0x55, 0x04, 0x07,
    0x0c, 0x09, 0x43, 0x75, 0x70, 0x65, 0x72, 0x74, 0x69, 0x6e, 0x6f, 0x82,
    0x01, 0x01, 0x30, 0x0c, 0x06, 0x03, 0x55, 0x1d, 0x13, 0x04, 0x05, 0x30,
    0x03, 0x01, 0x01, 0xff, 0x30, 0x0d, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86,
    0xf7, 0x0d, 0x01, 0x01, 0x05, 0x05, 0x00, 0x03, 0x81, 0x81, 0x00, 0x0b,
    0xea, 0xec, 0x19, 0xeb, 0x6f, 0x3d, 0x01, 0x28, 0x95, 0x40, 0x1a, 0x51,
    0xff, 0x62, 0x1e, 0xfd, 0xc7, 0x91, 0x61, 0x6f, 0x46, 0xca, 0xe1, 0x80,
    0xd7, 0x0e, 0x31, 0xf6, 0x16, 0xd1, 0x6b, 0x5e, 0x78, 0xdf, 0x02, 0xa2,
    0x8f, 0x35, 0x6f, 0x1e, 0x71, 0xd1, 0xd7, 0x4b, 0x5d, 0x4d, 0x7d, 0x0b,
    0x85, 0xd1, 0x7c, 0x4b, 0x84, 0x70, 0x22, 0xb3, 0xbd, 0x9e, 0x94, 0xaa,
    0x31, 0x31, 0x94, 0x81, 0x3b, 0x11, 0x03, 0x4d, 0x2a, 0xff, 0x81, 0xac,
    0xef, 0x95, 0x94, 0xf9, 0x0a, 0x73, 0xe9, 0xd0, 0x78, 0xbb, 0x65, 0xb1,
    0x5c, 0xcb, 0x1c, 0xff, 0xd4, 0x5b, 0x43, 0xf1, 0x12, 0x1b, 0xb0, 0xc1,
    0xa6, 0xb4, 0x7b, 0x82, 0x4c, 0x64, 0xa1, 0xad, 0x88, 0xe2, 0xe9, 0x89,
    0x62, 0xc0, 0x93, 0x8c, 0x0c, 0x42, 0x6c, 0xed, 0x12, 0x47, 0x16, 0x5c,
    0xba, 0xcf, 0x2f, 0x17, 0xdb, 0x07, 0x95
};

static
unsigned char serverD_cert_proof[] = {
    0x00, 0xab, 0xa8, 0xb5, 0xb4, 0x7d, 0x00, 0x00, 0x1b, 0x46, 0x58, 0x28,
    0xc4, 0x0a, 0xc7, 0x0b, 0x03, 0xf6, 0x91, 0x70, 0xa3, 0x5f, 0xed, 0xc8,
    0x74, 0x40, 0x3c, 0xd0, 0x58, 0x1d, 0x3c, 0x8c, 0x16, 0x00, 0x00, 0x01,
    0x47, 0xdc, 0x04, 0x70, 0x0e, 0x00, 0x00, 0x04, 0x03, 0x00, 0x46, 0x30,
    0x44, 0x02, 0x20, 0x71, 0x55, 0x2f, 0x75, 0xa8, 0x3a, 0xfd, 0x01, 0x34,
    0x44, 0xc7, 0x84, 0x71, 0x8f, 0x1e, 0xc2, 0x36, 0xe2, 0x08, 0x07, 0x92,
    0x2b, 0x9f, 0x44, 0x0e, 0x84, 0x16, 0x08, 0xe0, 0xaf, 0xc5, 0xb9, 0x02,
    0x20, 0x29, 0x3e, 0x0f, 0x63, 0x5c, 0xe7, 0x0a, 0xea, 0x1f, 0x96, 0x4a,
    0x11, 0x86, 0x72, 0x06, 0xa5, 0x25, 0xc4, 0x5e, 0xf6, 0x92, 0xd8, 0x08,
    0x98, 0x17, 0xba, 0xf2, 0xfe, 0x50, 0x62, 0x36, 0x29
};

static
unsigned char serverD_cert_der[] = {
    0x30, 0x82, 0x02, 0xe4, 0x30, 0x82, 0x02, 0x4d, 0xa0, 0x03, 0x02, 0x01,
    0x02, 0x02, 0x01, 0x13, 0x30, 0x0d, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86,
    0xf7, 0x0d, 0x01, 0x01, 0x05, 0x05, 0x00, 0x30, 0x52, 0x31, 0x0b, 0x30,
    0x09, 0x06, 0x03, 0x55, 0x04, 0x06, 0x13, 0x02, 0x55, 0x53, 0x31, 0x1a,
    0x30, 0x18, 0x06, 0x03, 0x55, 0x04, 0x0a, 0x0c, 0x11, 0x63, 0x6f, 0x72,
    0x65, 0x6f, 0x73, 0x2d, 0x63, 0x74, 0x2d, 0x74, 0x65, 0x73, 0x74, 0x20,
    0x43, 0x41, 0x31, 0x13, 0x30, 0x11, 0x06, 0x03, 0x55, 0x04, 0x08, 0x0c,
    0x0a, 0x43, 0x61, 0x6c, 0x69, 0x66, 0x6f, 0x72, 0x6e, 0x69, 0x61, 0x31,
    0x12, 0x30, 0x10, 0x06, 0x03, 0x55, 0x04, 0x07, 0x0c, 0x09, 0x43, 0x75,
    0x70, 0x65, 0x72, 0x74, 0x69, 0x6e, 0x6f, 0x30, 0x1e, 0x17, 0x0d, 0x31,
    0x32, 0x30, 0x36, 0x30, 0x31, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x5a,
    0x17, 0x0d, 0x32, 0x32, 0x30, 0x36, 0x30, 0x31, 0x30, 0x30, 0x30, 0x30,
    0x30, 0x30, 0x5a, 0x30, 0x72, 0x31, 0x0b, 0x30, 0x09, 0x06, 0x03, 0x55,
    0x04, 0x06, 0x13, 0x02, 0x55, 0x53, 0x31, 0x17, 0x30, 0x15, 0x06, 0x03,
    0x55, 0x04, 0x0a, 0x0c, 0x0e, 0x63, 0x6f, 0x72, 0x65, 0x6f, 0x73, 0x2d,
    0x63, 0x74, 0x2d, 0x74, 0x65, 0x73, 0x74, 0x31, 0x13, 0x30, 0x11, 0x06,
    0x03, 0x55, 0x04, 0x08, 0x0c, 0x0a, 0x43, 0x61, 0x6c, 0x69, 0x66, 0x6f,
    0x72, 0x6e, 0x69, 0x61, 0x31, 0x12, 0x30, 0x10, 0x06, 0x03, 0x55, 0x04,
    0x07, 0x0c, 0x09, 0x43, 0x75, 0x70, 0x65, 0x72, 0x74, 0x69, 0x6e, 0x6f,
    0x31, 0x21, 0x30, 0x1f, 0x06, 0x03, 0x55, 0x04, 0x03, 0x0c, 0x18, 0x63,
    0x6f, 0x72, 0x65, 0x6f, 0x73, 0x2d, 0x63, 0x74, 0x2d, 0x74, 0x65, 0x73,
    0x74, 0x2e, 0x61, 0x70, 0x70, 0x6c, 0x65, 0x2e, 0x63, 0x6f, 0x6d, 0x30,
    0x81, 0x9f, 0x30, 0x0d, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d,
    0x01, 0x01, 0x01, 0x05, 0x00, 0x03, 0x81, 0x8d, 0x00, 0x30, 0x81, 0x89,
    0x02, 0x81, 0x81, 0x00, 0xba, 0x00, 0xc4, 0xfb, 0x3f, 0x9a, 0x86, 0x43,
    0x1a, 0x26, 0x99, 0x9d, 0x19, 0x67, 0x27, 0xaa, 0x44, 0xd4, 0xba, 0x2b,
    0xfe, 0x7b, 0x32, 0xe8, 0x2a, 0xc7, 0x89, 0x36, 0x41, 0xd7, 0xaf, 0xf4,
    0x97, 0x4d, 0x41, 0x7b, 0xc7, 0x80, 0xba, 0x79, 0xab, 0x9c, 0xeb, 0xcc,
    0x38, 0xb7, 0x83, 0xdf, 0x62, 0x7e, 0xaf, 0x6c, 0x32, 0x57, 0xc2, 0x41,
    0xea, 0x73, 0xa9, 0x45, 0xf8, 0xbe, 0xc2, 0x26, 0x0f, 0x01, 0xec, 0x3b,
    0x02, 0x24, 0x7d, 0x39, 0x5c, 0xa6, 0x9c, 0xdf, 0x4b, 0x1f, 0xd5, 0x4d,
    0xd2, 0x5e, 0x9f, 0x09, 0x4c, 0x68, 0x11, 0xa3, 0x02, 0xb1, 0x65, 0x42,
    0xef, 0x67, 0x25, 0x30, 0x93, 0x86, 0x6f, 0x37, 0x1c, 0x83, 0x62, 0xd1,
    0x24, 0xfa, 0x89, 0x4d, 0x00, 0x8e, 0x77, 0x6a, 0xfd, 0x79, 0x85, 0x3e,
    0x59, 0xed, 0x92, 0xdf, 0x8a, 0xa1, 0xca, 0xfd, 0xfe, 0x1b, 0xf7, 0x1f,
    0x02, 0x03, 0x01, 0x00, 0x01, 0xa3, 0x81, 0xa9, 0x30, 0x81, 0xa6, 0x30,
    0x1d, 0x06, 0x03, 0x55, 0x1d, 0x0e, 0x04, 0x16, 0x04, 0x14, 0xf4, 0x42,
    0x90, 0xfd, 0x4c, 0xcd, 0x26, 0x10, 0x0b, 0xd7, 0x34, 0x22, 0xad, 0x23,
    0x26, 0xa0, 0x6c, 0xaf, 0xaa, 0x6c, 0x30, 0x7a, 0x06, 0x03, 0x55, 0x1d,
    0x23, 0x04, 0x73, 0x30, 0x71, 0x80, 0x14, 0xdc, 0x16, 0x44, 0x15, 0x3e,
    0x53, 0x27, 0xd8, 0x68, 0x66, 0x41, 0x40, 0x88, 0x90, 0xe4, 0x4e, 0x0a,
    0xda, 0x08, 0xa9, 0xa1, 0x56, 0xa4, 0x54, 0x30, 0x52, 0x31, 0x0b, 0x30,
    0x09, 0x06, 0x03, 0x55, 0x04, 0x06, 0x13, 0x02, 0x55, 0x53, 0x31, 0x1a,
    0x30, 0x18, 0x06, 0x03, 0x55, 0x04, 0x0a, 0x0c, 0x11, 0x63, 0x6f, 0x72,
    0x65, 0x6f, 0x73, 0x2d, 0x63, 0x74, 0x2d, 0x74, 0x65, 0x73, 0x74, 0x20,
    0x43, 0x41, 0x31, 0x13, 0x30, 0x11, 0x06, 0x03, 0x55, 0x04, 0x08, 0x0c,
    0x0a, 0x43, 0x61, 0x6c, 0x69, 0x66, 0x6f, 0x72, 0x6e, 0x69, 0x61, 0x31,
    0x12, 0x30, 0x10, 0x06, 0x03, 0x55, 0x04, 0x07, 0x0c, 0x09, 0x43, 0x75,
    0x70, 0x65, 0x72, 0x74, 0x69, 0x6e, 0x6f, 0x82, 0x01, 0x01, 0x30, 0x09,
    0x06, 0x03, 0x55, 0x1d, 0x13, 0x04, 0x02, 0x30, 0x00, 0x30, 0x0d, 0x06,
    0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01, 0x05, 0x05, 0x00,
    0x03, 0x81, 0x81, 0x00, 0x7a, 0x06, 0xe3, 0x17, 0xca, 0xee, 0xe0, 0x67,
    0x16, 0xfd, 0xf1, 0xad, 0x9f, 0xf8, 0xeb, 0xce, 0x03, 0x57, 0x7d, 0x90,
    0x6c, 0x85, 0xe0, 0x43, 0x3f, 0xb4, 0x3a, 0x08, 0x63, 0xef, 0x79, 0xf6,
    0xe1, 0xa3, 0x88, 0x32, 0xcf, 0x8f, 0x2f, 0xde, 0xd0, 0xc0, 0x92, 0x0b,
    0x16, 0xe1, 0xd4, 0x49, 0xd5, 0xb2, 0x84, 0x2e, 0x87, 0xfa, 0x1b, 0x5b,
    0x95, 0x51, 0x51, 0x0d, 0x29, 0x88, 0xd0, 0x8c, 0x10, 0x75, 0xe3, 0x78,
    0xb3, 0x4e, 0x39, 0xc1, 0xe4, 0xd0, 0x22, 0xb7, 0x64, 0xbe, 0xc3, 0x9d,
    0xff, 0x02, 0xc9, 0x66, 0xc3, 0x38, 0x4e, 0x88, 0xde, 0xa6, 0x75, 0x80,
    0xb3, 0x17, 0xb9, 0xfe, 0xfb, 0x64, 0xec, 0x3b, 0x16, 0xcd, 0xf0, 0x0d,
    0x15, 0xbf, 0x70, 0x42, 0xba, 0xe5, 0xec, 0x1d, 0x2f, 0xee, 0x0a, 0x2f,
    0xd7, 0x37, 0x9d, 0xc6, 0x0b, 0x26, 0xf3, 0xfb, 0x13, 0x69, 0x9f, 0x09
};

static
unsigned char serverF_cert_der[] = {
    0x30, 0x82, 0x03, 0x73, 0x30, 0x82, 0x02, 0xdc, 0xa0, 0x03, 0x02, 0x01,
    0x02, 0x02, 0x01, 0x15, 0x30, 0x0d, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86,
    0xf7, 0x0d, 0x01, 0x01, 0x05, 0x05, 0x00, 0x30, 0x52, 0x31, 0x0b, 0x30,
    0x09, 0x06, 0x03, 0x55, 0x04, 0x06, 0x13, 0x02, 0x55, 0x53, 0x31, 0x1a,
    0x30, 0x18, 0x06, 0x03, 0x55, 0x04, 0x0a, 0x0c, 0x11, 0x63, 0x6f, 0x72,
    0x65, 0x6f, 0x73, 0x2d, 0x63, 0x74, 0x2d, 0x74, 0x65, 0x73, 0x74, 0x20,
    0x43, 0x41, 0x31, 0x13, 0x30, 0x11, 0x06, 0x03, 0x55, 0x04, 0x08, 0x0c,
    0x0a, 0x43, 0x61, 0x6c, 0x69, 0x66, 0x6f, 0x72, 0x6e, 0x69, 0x61, 0x31,
    0x12, 0x30, 0x10, 0x06, 0x03, 0x55, 0x04, 0x07, 0x0c, 0x09, 0x43, 0x75,
    0x70, 0x65, 0x72, 0x74, 0x69, 0x6e, 0x6f, 0x30, 0x1e, 0x17, 0x0d, 0x31,
    0x32, 0x30, 0x36, 0x30, 0x31, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x5a,
    0x17, 0x0d, 0x32, 0x32, 0x30, 0x36, 0x30, 0x31, 0x30, 0x30, 0x30, 0x30,
    0x30, 0x30, 0x5a, 0x30, 0x72, 0x31, 0x0b, 0x30, 0x09, 0x06, 0x03, 0x55,
    0x04, 0x06, 0x13, 0x02, 0x55, 0x53, 0x31, 0x17, 0x30, 0x15, 0x06, 0x03,
    0x55, 0x04, 0x0a, 0x0c, 0x0e, 0x63, 0x6f, 0x72, 0x65, 0x6f, 0x73, 0x2d,
    0x63, 0x74, 0x2d, 0x74, 0x65, 0x73, 0x74, 0x31, 0x13, 0x30, 0x11, 0x06,
    0x03, 0x55, 0x04, 0x08, 0x0c, 0x0a, 0x43, 0x61, 0x6c, 0x69, 0x66, 0x6f,
    0x72, 0x6e, 0x69, 0x61, 0x31, 0x12, 0x30, 0x10, 0x06, 0x03, 0x55, 0x04,
    0x07, 0x0c, 0x09, 0x43, 0x75, 0x70, 0x65, 0x72, 0x74, 0x69, 0x6e, 0x6f,
    0x31, 0x21, 0x30, 0x1f, 0x06, 0x03, 0x55, 0x04, 0x03, 0x0c, 0x18, 0x63,
    0x6f, 0x72, 0x65, 0x6f, 0x73, 0x2d, 0x63, 0x74, 0x2d, 0x74, 0x65, 0x73,
    0x74, 0x2e, 0x61, 0x70, 0x70, 0x6c, 0x65, 0x2e, 0x63, 0x6f, 0x6d, 0x30,
    0x81, 0x9f, 0x30, 0x0d, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d,
    0x01, 0x01, 0x01, 0x05, 0x00, 0x03, 0x81, 0x8d, 0x00, 0x30, 0x81, 0x89,
    0x02, 0x81, 0x81, 0x00, 0xd2, 0x49, 0x0c, 0xd0, 0xc5, 0xa8, 0xc3, 0x0f,
    0x36, 0x99, 0x54, 0x00, 0xd7, 0xf0, 0x2a, 0xcb, 0x21, 0x20, 0x4c, 0xac,
    0xaa, 0xcb, 0x36, 0x20, 0x72, 0x78, 0x05, 0xd1, 0xc2, 0xf9, 0xce, 0xc9,
    0x5b, 0xbc, 0x38, 0xda, 0xdd, 0x27, 0xf7, 0x6b, 0x0a, 0xf0, 0x16, 0xe2,
    0xc9, 0x74, 0x8c, 0x47, 0x5b, 0x07, 0x91, 0xa5, 0x6c, 0xcf, 0xf9, 0x0a,
    0x05, 0xb3, 0x05, 0x6a, 0xbe, 0x59, 0xdb, 0xa2, 0x1b, 0x21, 0x29, 0xe1,
    0xef, 0x0d, 0x4f, 0xa1, 0xc5, 0xbd, 0x16, 0xeb, 0x8c, 0x45, 0x6f, 0x64,
    0x42, 0x93, 0x82, 0xb3, 0x6d, 0xff, 0x83, 0x61, 0xdc, 0xcf, 0x8d, 0xd0,
    0x09, 0x2c, 0x37, 0x87, 0x1b, 0x75, 0xf6, 0xb3, 0xf8, 0x45, 0xef, 0xe2,
    0xcb, 0xff, 0x6d, 0xbb, 0xe4, 0xa5, 0x29, 0xee, 0xc0, 0x78, 0x17, 0x94,
    0xdc, 0x6b, 0xc7, 0x46, 0x01, 0x74, 0xf9, 0x65, 0x3b, 0x59, 0x21, 0xf5,
    0x02, 0x03, 0x01, 0x00, 0x01, 0xa3, 0x82, 0x01, 0x37, 0x30, 0x82, 0x01,
    0x33, 0x30, 0x1d, 0x06, 0x03, 0x55, 0x1d, 0x0e, 0x04, 0x16, 0x04, 0x14,
    0x69, 0x9d, 0x9f, 0x7e, 0xd9, 0x34, 0x7c, 0xfa, 0xd5, 0xc2, 0x7e, 0x02,
    0x0f, 0x1e, 0x4d, 0x1d, 0xa9, 0x8e, 0xa8, 0xcb, 0x30, 0x7a, 0x06, 0x03,
    0x55, 0x1d, 0x23, 0x04, 0x73, 0x30, 0x71, 0x80, 0x14, 0xdc, 0x16, 0x44,
    0x15, 0x3e, 0x53, 0x27, 0xd8, 0x68, 0x66, 0x41, 0x40, 0x88, 0x90, 0xe4,
    0x4e, 0x0a, 0xda, 0x08, 0xa9, 0xa1, 0x56, 0xa4, 0x54, 0x30, 0x52, 0x31,
    0x0b, 0x30, 0x09, 0x06, 0x03, 0x55, 0x04, 0x06, 0x13, 0x02, 0x55, 0x53,
    0x31, 0x1a, 0x30, 0x18, 0x06, 0x03, 0x55, 0x04, 0x0a, 0x0c, 0x11, 0x63,
    0x6f, 0x72, 0x65, 0x6f, 0x73, 0x2d, 0x63, 0x74, 0x2d, 0x74, 0x65, 0x73,
    0x74, 0x20, 0x43, 0x41, 0x31, 0x13, 0x30, 0x11, 0x06, 0x03, 0x55, 0x04,
    0x08, 0x0c, 0x0a, 0x43, 0x61, 0x6c, 0x69, 0x66, 0x6f, 0x72, 0x6e, 0x69,
    0x61, 0x31, 0x12, 0x30, 0x10, 0x06, 0x03, 0x55, 0x04, 0x07, 0x0c, 0x09,
    0x43, 0x75, 0x70, 0x65, 0x72, 0x74, 0x69, 0x6e, 0x6f, 0x82, 0x01, 0x01,
    0x30, 0x09, 0x06, 0x03, 0x55, 0x1d, 0x13, 0x04, 0x02, 0x30, 0x00, 0x30,
    0x81, 0x8a, 0x06, 0x0a, 0x2b, 0x06, 0x01, 0x04, 0x01, 0xd6, 0x79, 0x02,
    0x04, 0x02, 0x04, 0x7c, 0x04, 0x7a, 0x00, 0x78, 0x00, 0x76, 0x00, 0xab,
    0xa8, 0xb5, 0xb4, 0x7d, 0x00, 0x00, 0x1b, 0x46, 0x58, 0x28, 0xc4, 0x0a,
    0xc7, 0x0b, 0x03, 0xf6, 0x91, 0x70, 0xa3, 0x5f, 0xed, 0xc8, 0x74, 0x40,
    0x3c, 0xd0, 0x58, 0x1d, 0x3c, 0x8c, 0x16, 0x00, 0x00, 0x01, 0x47, 0xdc,
    0x04, 0xbc, 0x5a, 0x00, 0x00, 0x04, 0x03, 0x00, 0x47, 0x30, 0x45, 0x02,
    0x20, 0x5b, 0x3b, 0xe2, 0x6b, 0xa2, 0xda, 0x49, 0xb2, 0xa5, 0x55, 0x1d,
    0x2f, 0x4d, 0x21, 0x2e, 0x2d, 0xf7, 0x59, 0xb3, 0x22, 0x1d, 0x90, 0x38,
    0x88, 0x77, 0xad, 0x49, 0xca, 0x28, 0x1d, 0x4a, 0xa8, 0x02, 0x21, 0x00,
    0xb7, 0x08, 0x08, 0xfb, 0x6a, 0x06, 0x13, 0xaa, 0xe6, 0x4d, 0x69, 0x44,
    0xce, 0xc0, 0x17, 0x8f, 0x3e, 0x80, 0x30, 0xe2, 0xd0, 0xe1, 0x8b, 0xc0,
    0x34, 0x28, 0x8b, 0xd8, 0x85, 0xb5, 0x14, 0x97, 0x30, 0x0d, 0x06, 0x09,
    0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01, 0x05, 0x05, 0x00, 0x03,
    0x81, 0x81, 0x00, 0xae, 0x8c, 0x7f, 0x63, 0x9d, 0xdd, 0xee, 0x4f, 0xc4,
    0xc5, 0x7b, 0x20, 0xb5, 0xe8, 0x89, 0x3e, 0x2c, 0xfe, 0x36, 0x0e, 0x31,
    0x1a, 0x38, 0xd6, 0xb3, 0xfd, 0x37, 0xeb, 0x26, 0xd0, 0x27, 0xfa, 0x04,
    0x12, 0x9c, 0xe2, 0x20, 0xe1, 0x61, 0xbf, 0xee, 0x60, 0x45, 0x84, 0xa0,
    0xea, 0xce, 0x1f, 0xf7, 0x73, 0x31, 0xd4, 0xd7, 0x87, 0xe7, 0xd5, 0x9f,
    0xff, 0x8d, 0x14, 0x32, 0x22, 0x89, 0xf6, 0x31, 0x38, 0xef, 0x1c, 0x36,
    0x55, 0x0d, 0x5f, 0x0d, 0x99, 0x36, 0x58, 0x6a, 0xa3, 0xff, 0xf0, 0xc7,
    0xe0, 0x5e, 0x02, 0x20, 0x9f, 0x04, 0x0a, 0xa4, 0xba, 0x1a, 0x1c, 0xb2,
    0x43, 0x85, 0xc2, 0xcc, 0xd2, 0x95, 0x8f, 0x20, 0x11, 0x1d, 0xea, 0x9e,
    0x10, 0xf1, 0x45, 0xd2, 0x4d, 0x95, 0x80, 0xed, 0xe1, 0x86, 0x71, 0xee,
    0x50, 0x0f, 0xb0, 0x73, 0x12, 0x32, 0xdd, 0x95, 0xc5, 0xb9, 0x54
};


static uint8_t server_A_proof_Alfa_3[] = {
    0x00, 0xcb, 0xbd, 0xa5, 0xa5, 0x4f, 0x94, 0xc5, 0xa5, 0x19, 0x3e, 0xcc,
    0x98, 0xa4, 0x8f, 0x93, 0x39, 0xf3, 0x01, 0x4d, 0x89, 0xbb, 0x41, 0x38,
    0x88, 0x49, 0x64, 0x91, 0x23, 0xac, 0x06, 0xcf, 0x8f, 0x00, 0x00, 0x01,
    0x4b, 0xf1, 0xa3, 0x17, 0xe0, 0x00, 0x00, 0x04, 0x03, 0x00, 0x47, 0x30,
    0x45, 0x02, 0x21, 0x00, 0x8a, 0x3f, 0xa6, 0x88, 0x76, 0x1b, 0x7b, 0x9b,
    0x30, 0x9b, 0x2d, 0x54, 0xe7, 0xe3, 0x1a, 0xda, 0xf6, 0xad, 0x76, 0xba,
    0x5a, 0x87, 0x24, 0x18, 0xc6, 0xdf, 0x93, 0xe7, 0x09, 0x67, 0xcb, 0x76,
    0x02, 0x20, 0x55, 0x59, 0xe2, 0x67, 0xe2, 0x9b, 0x9c, 0x77, 0x8e, 0x59,
    0x4c, 0xe8, 0x71, 0x69, 0x6d, 0x9a, 0x52, 0xbd, 0xea, 0xfb, 0xd7, 0x36,
    0x9e, 0xd0, 0x19, 0x95, 0x26, 0xb4, 0x57, 0xf8, 0x76, 0x6c
};

static uint8_t server_A_proof_Bravo_3[] = {
    0x00, 0x3b, 0x1a, 0xe4, 0xe3, 0xf9, 0x15, 0xa5, 0xa2, 0x77, 0xfe, 0x00,
    0xcf, 0x5d, 0xf1, 0x03, 0x50, 0xf9, 0x0d, 0x0b, 0x0e, 0x6d, 0x4a, 0x6f,
    0xa0, 0xcf, 0xe9, 0xa3, 0x76, 0xc0, 0xd2, 0x47, 0x4a, 0x00, 0x00, 0x01,
    0x4b, 0xf1, 0xa3, 0x1f, 0xd3, 0x00, 0x00, 0x04, 0x03, 0x00, 0x47, 0x30,
    0x45, 0x02, 0x21, 0x00, 0xc0, 0x8f, 0xd0, 0xf9, 0x9b, 0x8a, 0xb4, 0xa2,
    0x89, 0x7f, 0xc0, 0x86, 0xe4, 0xc2, 0x95, 0xa1, 0xcf, 0xaa, 0xd5, 0xff,
    0xe9, 0xdc, 0x46, 0xd1, 0xda, 0xf7, 0xe7, 0x9f, 0x8a, 0x9e, 0x6f, 0x5b,
    0x02, 0x20, 0x5c, 0x80, 0xd3, 0x25, 0x48, 0xb4, 0xa7, 0x97, 0x47, 0xcb,
    0x90, 0x9e, 0x56, 0xf3, 0xea, 0x66, 0x62, 0x0c, 0x7c, 0xb5, 0x49, 0xb1,
    0x12, 0xe9, 0xf9, 0x76, 0x8c, 0x70, 0xd7, 0x55, 0xbc, 0x89
};


static uint8_t server_A_cert_der[] = {
    0x30, 0x82, 0x02, 0xe4, 0x30, 0x82, 0x02, 0x4d, 0xa0, 0x03, 0x02, 0x01,
    0x02, 0x02, 0x01, 0x4b, 0x30, 0x0d, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86,
    0xf7, 0x0d, 0x01, 0x01, 0x05, 0x05, 0x00, 0x30, 0x52, 0x31, 0x0b, 0x30,
    0x09, 0x06, 0x03, 0x55, 0x04, 0x06, 0x13, 0x02, 0x55, 0x53, 0x31, 0x1a,
    0x30, 0x18, 0x06, 0x03, 0x55, 0x04, 0x0a, 0x0c, 0x11, 0x63, 0x6f, 0x72,
    0x65, 0x6f, 0x73, 0x2d, 0x63, 0x74, 0x2d, 0x74, 0x65, 0x73, 0x74, 0x20,
    0x43, 0x41, 0x31, 0x13, 0x30, 0x11, 0x06, 0x03, 0x55, 0x04, 0x08, 0x0c,
    0x0a, 0x43, 0x61, 0x6c, 0x69, 0x66, 0x6f, 0x72, 0x6e, 0x69, 0x61, 0x31,
    0x12, 0x30, 0x10, 0x06, 0x03, 0x55, 0x04, 0x07, 0x0c, 0x09, 0x43, 0x75,
    0x70, 0x65, 0x72, 0x74, 0x69, 0x6e, 0x6f, 0x30, 0x1e, 0x17, 0x0d, 0x31,
    0x35, 0x30, 0x31, 0x30, 0x31, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x5a,
    0x17, 0x0d, 0x31, 0x36, 0x30, 0x31, 0x30, 0x31, 0x30, 0x30, 0x30, 0x30,
    0x30, 0x30, 0x5a, 0x30, 0x72, 0x31, 0x0b, 0x30, 0x09, 0x06, 0x03, 0x55,
    0x04, 0x06, 0x13, 0x02, 0x55, 0x53, 0x31, 0x17, 0x30, 0x15, 0x06, 0x03,
    0x55, 0x04, 0x0a, 0x0c, 0x0e, 0x63, 0x6f, 0x72, 0x65, 0x6f, 0x73, 0x2d,
    0x63, 0x74, 0x2d, 0x74, 0x65, 0x73, 0x74, 0x31, 0x13, 0x30, 0x11, 0x06,
    0x03, 0x55, 0x04, 0x08, 0x0c, 0x0a, 0x43, 0x61, 0x6c, 0x69, 0x66, 0x6f,
    0x72, 0x6e, 0x69, 0x61, 0x31, 0x12, 0x30, 0x10, 0x06, 0x03, 0x55, 0x04,
    0x07, 0x0c, 0x09, 0x43, 0x75, 0x70, 0x65, 0x72, 0x74, 0x69, 0x6e, 0x6f,
    0x31, 0x21, 0x30, 0x1f, 0x06, 0x03, 0x55, 0x04, 0x03, 0x0c, 0x18, 0x63,
    0x6f, 0x72, 0x65, 0x6f, 0x73, 0x2d, 0x63, 0x74, 0x2d, 0x74, 0x65, 0x73,
    0x74, 0x2e, 0x61, 0x70, 0x70, 0x6c, 0x65, 0x2e, 0x63, 0x6f, 0x6d, 0x30,
    0x81, 0x9f, 0x30, 0x0d, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d,
    0x01, 0x01, 0x01, 0x05, 0x00, 0x03, 0x81, 0x8d, 0x00, 0x30, 0x81, 0x89,
    0x02, 0x81, 0x81, 0x00, 0xc2, 0x65, 0x91, 0x0d, 0xf6, 0x8a, 0x40, 0x46,
    0xbd, 0x9e, 0x90, 0xdb, 0xf8, 0x12, 0x1c, 0xfa, 0x76, 0xe4, 0x7b, 0x14,
    0xf5, 0x3f, 0xef, 0xf7, 0x5f, 0x34, 0x55, 0xf7, 0x9e, 0x59, 0xeb, 0x48,
    0xac, 0xb6, 0x40, 0x77, 0xb9, 0x0a, 0x64, 0xc3, 0xe8, 0xdd, 0xbd, 0x52,
    0x24, 0x4f, 0xb7, 0x22, 0xed, 0xe4, 0xc8, 0xac, 0x9e, 0x9e, 0x2d, 0xe1,
    0x66, 0xa9, 0x40, 0x56, 0xdb, 0x0a, 0xb9, 0x3a, 0x69, 0xd2, 0xcf, 0x3c,
    0xfa, 0x17, 0x44, 0xc7, 0x4f, 0xc0, 0xf8, 0xba, 0x20, 0x68, 0xc6, 0x75,
    0x38, 0xa0, 0xc5, 0xd4, 0x1a, 0x5c, 0x86, 0xbb, 0x95, 0xa8, 0x71, 0x3e,
    0xfc, 0xf3, 0xb6, 0x74, 0x92, 0x98, 0x21, 0xec, 0x03, 0x90, 0x97, 0x71,
    0xa2, 0xd5, 0x79, 0xcf, 0x2a, 0x59, 0xcf, 0x16, 0xdc, 0x0b, 0x03, 0x9d,
    0xfd, 0x60, 0xad, 0x5f, 0x7f, 0xa5, 0x0a, 0x24, 0x9b, 0x83, 0xc5, 0x63,
    0x02, 0x03, 0x01, 0x00, 0x01, 0xa3, 0x81, 0xa9, 0x30, 0x81, 0xa6, 0x30,
    0x1d, 0x06, 0x03, 0x55, 0x1d, 0x0e, 0x04, 0x16, 0x04, 0x14, 0x8d, 0x37,
    0x29, 0x3a, 0x88, 0xa5, 0x5c, 0x19, 0x3b, 0xd1, 0x21, 0x37, 0xf3, 0xed,
    0xd5, 0x2a, 0xf5, 0xb3, 0x49, 0x87, 0x30, 0x7a, 0x06, 0x03, 0x55, 0x1d,
    0x23, 0x04, 0x73, 0x30, 0x71, 0x80, 0x14, 0x64, 0x1f, 0x09, 0x99, 0x2d,
    0x6a, 0x5b, 0x4d, 0xef, 0xed, 0xbb, 0xba, 0x96, 0xf9, 0x73, 0x65, 0xad,
    0x6e, 0x84, 0xbd, 0xa1, 0x56, 0xa4, 0x54, 0x30, 0x52, 0x31, 0x0b, 0x30,
    0x09, 0x06, 0x03, 0x55, 0x04, 0x06, 0x13, 0x02, 0x55, 0x53, 0x31, 0x1a,
    0x30, 0x18, 0x06, 0x03, 0x55, 0x04, 0x0a, 0x0c, 0x11, 0x63, 0x6f, 0x72,
    0x65, 0x6f, 0x73, 0x2d, 0x63, 0x74, 0x2d, 0x74, 0x65, 0x73, 0x74, 0x20,
    0x43, 0x41, 0x31, 0x13, 0x30, 0x11, 0x06, 0x03, 0x55, 0x04, 0x08, 0x0c,
    0x0a, 0x43, 0x61, 0x6c, 0x69, 0x66, 0x6f, 0x72, 0x6e, 0x69, 0x61, 0x31,
    0x12, 0x30, 0x10, 0x06, 0x03, 0x55, 0x04, 0x07, 0x0c, 0x09, 0x43, 0x75,
    0x70, 0x65, 0x72, 0x74, 0x69, 0x6e, 0x6f, 0x82, 0x01, 0x01, 0x30, 0x09,
    0x06, 0x03, 0x55, 0x1d, 0x13, 0x04, 0x02, 0x30, 0x00, 0x30, 0x0d, 0x06,
    0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01, 0x05, 0x05, 0x00,
    0x03, 0x81, 0x81, 0x00, 0x27, 0xe3, 0x99, 0x5e, 0xea, 0x13, 0xa4, 0x92,
    0xc3, 0xe4, 0xa4, 0xf9, 0x8c, 0xf0, 0x41, 0x49, 0x49, 0x40, 0x51, 0xba,
    0x14, 0xe6, 0xa4, 0x5b, 0x84, 0xb7, 0xc3, 0x57, 0x39, 0xa2, 0x0b, 0xd0,
    0x20, 0x66, 0xd2, 0x91, 0xa9, 0xd1, 0x14, 0x2f, 0x7f, 0xe1, 0x59, 0x5f,
    0xff, 0x37, 0x9e, 0xfb, 0xed, 0xea, 0xc5, 0x79, 0x1f, 0x34, 0xcc, 0x15,
    0xf0, 0xc4, 0x01, 0x06, 0x12, 0x5a, 0x07, 0x81, 0xaf, 0x6c, 0x99, 0x72,
    0x5a, 0x6c, 0x0f, 0x89, 0xde, 0x01, 0x48, 0xf6, 0xa5, 0x12, 0x0e, 0x6b,
    0xd6, 0x13, 0x40, 0x9b, 0xd1, 0x9f, 0xb5, 0x39, 0x49, 0x07, 0x2d, 0x04,
    0x34, 0xe1, 0x04, 0x83, 0xa4, 0x74, 0x0c, 0x52, 0x54, 0x93, 0x6a, 0x63,
    0x05, 0x06, 0xbf, 0x1d, 0x01, 0x30, 0xab, 0xf0, 0xed, 0x46, 0x26, 0x75,
    0x5e, 0x9a, 0xa4, 0x01, 0xe6, 0x95, 0x65, 0xc5, 0xb9, 0x09, 0x84, 0x98
};

static
unsigned char www_digicert_com_cert_der[] = {
    0x30, 0x82, 0x08, 0x5c, 0x30, 0x82, 0x07, 0x44, 0xa0, 0x03, 0x02, 0x01,
    0x02, 0x02, 0x10, 0x04, 0xc3, 0x34, 0xf6, 0xf8, 0x98, 0x9b, 0xb5, 0x4c,
    0xb1, 0x91, 0x94, 0x30, 0x53, 0x88, 0xb7, 0x30, 0x0d, 0x06, 0x09, 0x2a,
    0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01, 0x0b, 0x05, 0x00, 0x30, 0x75,
    0x31, 0x0b, 0x30, 0x09, 0x06, 0x03, 0x55, 0x04, 0x06, 0x13, 0x02, 0x55,
    0x53, 0x31, 0x15, 0x30, 0x13, 0x06, 0x03, 0x55, 0x04, 0x0a, 0x13, 0x0c,
    0x44, 0x69, 0x67, 0x69, 0x43, 0x65, 0x72, 0x74, 0x20, 0x49, 0x6e, 0x63,
    0x31, 0x19, 0x30, 0x17, 0x06, 0x03, 0x55, 0x04, 0x0b, 0x13, 0x10, 0x77,
    0x77, 0x77, 0x2e, 0x64, 0x69, 0x67, 0x69, 0x63, 0x65, 0x72, 0x74, 0x2e,
    0x63, 0x6f, 0x6d, 0x31, 0x34, 0x30, 0x32, 0x06, 0x03, 0x55, 0x04, 0x03,
    0x13, 0x2b, 0x44, 0x69, 0x67, 0x69, 0x43, 0x65, 0x72, 0x74, 0x20, 0x53,
    0x48, 0x41, 0x32, 0x20, 0x45, 0x78, 0x74, 0x65, 0x6e, 0x64, 0x65, 0x64,
    0x20, 0x56, 0x61, 0x6c, 0x69, 0x64, 0x61, 0x74, 0x69, 0x6f, 0x6e, 0x20,
    0x53, 0x65, 0x72, 0x76, 0x65, 0x72, 0x20, 0x43, 0x41, 0x30, 0x1e, 0x17,
    0x0d, 0x31, 0x34, 0x31, 0x30, 0x31, 0x34, 0x30, 0x30, 0x30, 0x30, 0x30,
    0x30, 0x5a, 0x17, 0x0d, 0x31, 0x36, 0x31, 0x30, 0x31, 0x38, 0x31, 0x32,
    0x30, 0x30, 0x30, 0x30, 0x5a, 0x30, 0x82, 0x01, 0x0b, 0x31, 0x1d, 0x30,
    0x1b, 0x06, 0x03, 0x55, 0x04, 0x0f, 0x0c, 0x14, 0x50, 0x72, 0x69, 0x76,
    0x61, 0x74, 0x65, 0x20, 0x4f, 0x72, 0x67, 0x61, 0x6e, 0x69, 0x7a, 0x61,
    0x74, 0x69, 0x6f, 0x6e, 0x31, 0x13, 0x30, 0x11, 0x06, 0x0b, 0x2b, 0x06,
    0x01, 0x04, 0x01, 0x82, 0x37, 0x3c, 0x02, 0x01, 0x03, 0x13, 0x02, 0x55,
    0x53, 0x31, 0x15, 0x30, 0x13, 0x06, 0x0b, 0x2b, 0x06, 0x01, 0x04, 0x01,
    0x82, 0x37, 0x3c, 0x02, 0x01, 0x02, 0x13, 0x04, 0x55, 0x74, 0x61, 0x68,
    0x31, 0x15, 0x30, 0x13, 0x06, 0x03, 0x55, 0x04, 0x05, 0x13, 0x0c, 0x35,
    0x32, 0x39, 0x39, 0x35, 0x33, 0x37, 0x2d, 0x30, 0x31, 0x34, 0x32, 0x31,
    0x12, 0x30, 0x10, 0x06, 0x03, 0x55, 0x04, 0x09, 0x13, 0x09, 0x53, 0x75,
    0x69, 0x74, 0x65, 0x20, 0x35, 0x30, 0x30, 0x31, 0x24, 0x30, 0x22, 0x06,
    0x03, 0x55, 0x04, 0x09, 0x13, 0x1b, 0x32, 0x36, 0x30, 0x30, 0x20, 0x57,
    0x65, 0x73, 0x74, 0x20, 0x45, 0x78, 0x65, 0x63, 0x75, 0x74, 0x69, 0x76,
    0x65, 0x20, 0x50, 0x61, 0x72, 0x6b, 0x77, 0x61, 0x79, 0x31, 0x0e, 0x30,
    0x0c, 0x06, 0x03, 0x55, 0x04, 0x11, 0x13, 0x05, 0x38, 0x34, 0x30, 0x34,
    0x33, 0x31, 0x0b, 0x30, 0x09, 0x06, 0x03, 0x55, 0x04, 0x06, 0x13, 0x02,
    0x55, 0x53, 0x31, 0x0d, 0x30, 0x0b, 0x06, 0x03, 0x55, 0x04, 0x08, 0x13,
    0x04, 0x55, 0x74, 0x61, 0x68, 0x31, 0x0d, 0x30, 0x0b, 0x06, 0x03, 0x55,
    0x04, 0x07, 0x13, 0x04, 0x4c, 0x65, 0x68, 0x69, 0x31, 0x17, 0x30, 0x15,
    0x06, 0x03, 0x55, 0x04, 0x0a, 0x13, 0x0e, 0x44, 0x69, 0x67, 0x69, 0x43,
    0x65, 0x72, 0x74, 0x2c, 0x20, 0x49, 0x6e, 0x63, 0x2e, 0x31, 0x19, 0x30,
    0x17, 0x06, 0x03, 0x55, 0x04, 0x03, 0x13, 0x10, 0x77, 0x77, 0x77, 0x2e,
    0x64, 0x69, 0x67, 0x69, 0x63, 0x65, 0x72, 0x74, 0x2e, 0x63, 0x6f, 0x6d,
    0x30, 0x82, 0x02, 0x22, 0x30, 0x0d, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86,
    0xf7, 0x0d, 0x01, 0x01, 0x01, 0x05, 0x00, 0x03, 0x82, 0x02, 0x0f, 0x00,
    0x30, 0x82, 0x02, 0x0a, 0x02, 0x82, 0x02, 0x01, 0x00, 0xa8, 0x89, 0xb3,
    0x3b, 0x91, 0x94, 0x57, 0x87, 0x72, 0x09, 0x5b, 0x5f, 0xcb, 0x2c, 0x42,
    0x2a, 0x9e, 0xed, 0xc2, 0xfd, 0x20, 0x7b, 0x2c, 0x63, 0x7f, 0xdd, 0x07,
    0xbf, 0xfb, 0x49, 0x5c, 0xed, 0x1c, 0xa2, 0x70, 0x79, 0x75, 0xc2, 0x34,
    0xcc, 0xeb, 0x12, 0xf0, 0x40, 0x88, 0x3a, 0xb9, 0xea, 0x29, 0xa2, 0x11,
    0x8f, 0x53, 0xe1, 0x02, 0xe1, 0x87, 0x04, 0xf6, 0x58, 0xb9, 0x86, 0xb6,
    0x7f, 0x85, 0x5e, 0x0a, 0x58, 0x47, 0xc3, 0xbd, 0xe7, 0x6b, 0x21, 0x07,
    0x9d, 0xdb, 0xef, 0x57, 0x8b, 0x16, 0xce, 0x38, 0xf1, 0xe3, 0xe2, 0xe4,
    0x5a, 0x10, 0xb8, 0x39, 0xbb, 0x0a, 0xad, 0xca, 0xc5, 0x10, 0x85, 0x3a,
    0xa1, 0x6f, 0x67, 0xc9, 0x18, 0xc3, 0x5b, 0xb2, 0x4c, 0xa6, 0x01, 0xb6,
    0xc3, 0x50, 0xbe, 0x7e, 0xc8, 0x79, 0xca, 0x3c, 0x53, 0x5e, 0x02, 0x78,
    0xae, 0x96, 0x5f, 0x56, 0x21, 0xb3, 0xa4, 0x3c, 0x3f, 0xfe, 0x49, 0xc5,
    0x17, 0x73, 0xa5, 0x6e, 0xa9, 0x60, 0xaa, 0xbd, 0x16, 0x04, 0x56, 0xfa,
    0x54, 0xd2, 0xcb, 0x25, 0xc0, 0xe9, 0x9f, 0x89, 0xc9, 0xee, 0x10, 0x87,
    0x01, 0xf2, 0xc7, 0x93, 0x2d, 0xc3, 0x2f, 0x9e, 0xd0, 0x9c, 0x42, 0x24,
    0x9d, 0x09, 0x24, 0xf6, 0x80, 0xc4, 0xe8, 0x34, 0x99, 0x5a, 0x2e, 0x26,
    0xc3, 0x73, 0x28, 0x52, 0x26, 0xac, 0x09, 0x34, 0x8e, 0xc5, 0x70, 0xe1,
    0xf5, 0xfb, 0x93, 0xb8, 0x34, 0x2d, 0x44, 0xf4, 0x50, 0x1f, 0x86, 0x0a,
    0x9b, 0x64, 0x45, 0x26, 0x05, 0xd4, 0x45, 0xca, 0x72, 0x03, 0xdd, 0x1e,
    0x80, 0x1a, 0x9c, 0x53, 0x06, 0x7b, 0xc8, 0x36, 0x31, 0x03, 0xda, 0x5f,
    0x55, 0xc4, 0x0d, 0x29, 0xc0, 0x52, 0x9c, 0x23, 0x95, 0x8d, 0xa9, 0x55,
    0x95, 0xc4, 0x11, 0x02, 0x5b, 0xa3, 0x1b, 0xee, 0x79, 0xb2, 0x6e, 0x4a,
    0x6a, 0x4d, 0x4a, 0x44, 0x3e, 0x39, 0x9e, 0x8b, 0x0d, 0xec, 0x38, 0x93,
    0x5e, 0x5c, 0xb3, 0x4f, 0x53, 0x8f, 0x4e, 0x2a, 0x78, 0xb1, 0x52, 0x54,
    0x4b, 0xfb, 0x6a, 0x94, 0x35, 0x61, 0x03, 0x06, 0x79, 0xe8, 0x06, 0x9c,
    0x8e, 0x81, 0x5b, 0x6b, 0x36, 0xdf, 0xc0, 0xfe, 0x43, 0xce, 0xd5, 0x16,
    0x19, 0xf6, 0x82, 0x94, 0xe8, 0x80, 0x00, 0xe1, 0x84, 0x14, 0x1d, 0x28,
    0x73, 0x8b, 0xe9, 0xba, 0xb6, 0x55, 0xe7, 0xa6, 0x17, 0x8c, 0xae, 0x70,
    0x15, 0xbe, 0x04, 0xef, 0xc8, 0x08, 0x27, 0xd9, 0xdf, 0x3a, 0x7e, 0x67,
    0x8c, 0x06, 0x0d, 0x51, 0x94, 0x05, 0x95, 0x2f, 0x27, 0xe4, 0xc1, 0xd4,
    0xa4, 0x5e, 0xca, 0x96, 0x13, 0x89, 0xd2, 0x05, 0x8b, 0x43, 0x68, 0xfc,
    0x31, 0x87, 0xa9, 0xb6, 0xf2, 0xc3, 0x47, 0xe3, 0xdf, 0xd9, 0x19, 0x13,
    0x4f, 0xb9, 0x05, 0xa9, 0x8a, 0x98, 0x03, 0xca, 0xc5, 0x92, 0x29, 0xe3,
    0x73, 0xe7, 0x4b, 0xe8, 0x0a, 0xda, 0x1b, 0x9c, 0xdb, 0x68, 0x50, 0x66,
    0x95, 0x2b, 0xdc, 0xe8, 0x39, 0x1b, 0x14, 0xfa, 0x41, 0xd3, 0xfc, 0xda,
    0xe6, 0x8d, 0x04, 0x2c, 0x81, 0xd1, 0x12, 0x47, 0xc6, 0x27, 0x9d, 0xd7,
    0x54, 0xbd, 0x4f, 0xee, 0x42, 0x20, 0x96, 0x52, 0xa6, 0x83, 0x9f, 0x59,
    0x05, 0x6b, 0x2b, 0x18, 0x41, 0x7a, 0x5a, 0xbb, 0x89, 0x1b, 0x45, 0x82,
    0x8a, 0x6e, 0x7b, 0x94, 0x78, 0xe0, 0x4e, 0x09, 0xeb, 0x1c, 0xa8, 0xda,
    0xd9, 0xb4, 0x56, 0xd4, 0xa0, 0x7d, 0x08, 0xd5, 0xf2, 0x94, 0x81, 0x2e,
    0xa1, 0xb4, 0x0a, 0x14, 0x56, 0x21, 0x26, 0xc3, 0xc4, 0x27, 0x48, 0x3c,
    0x50, 0xd5, 0x71, 0x45, 0x35, 0x4b, 0x37, 0x22, 0x7b, 0x69, 0x26, 0x6c,
    0xdb, 0xb8, 0x4e, 0xf2, 0xf1, 0xa2, 0xf8, 0x6b, 0xfb, 0x1a, 0xae, 0xe6,
    0xeb, 0x5b, 0x1e, 0x15, 0xd5, 0x02, 0x03, 0x01, 0x00, 0x01, 0xa3, 0x82,
    0x03, 0x4e, 0x30, 0x82, 0x03, 0x4a, 0x30, 0x1f, 0x06, 0x03, 0x55, 0x1d,
    0x23, 0x04, 0x18, 0x30, 0x16, 0x80, 0x14, 0x3d, 0xd3, 0x50, 0xa5, 0xd6,
    0xa0, 0xad, 0xee, 0xf3, 0x4a, 0x60, 0x0a, 0x65, 0xd3, 0x21, 0xd4, 0xf8,
    0xf8, 0xd6, 0x0f, 0x30, 0x1d, 0x06, 0x03, 0x55, 0x1d, 0x0e, 0x04, 0x16,
    0x04, 0x14, 0xf8, 0xa3, 0xa7, 0x61, 0xab, 0xd9, 0x77, 0x4b, 0x19, 0x66,
    0x90, 0xc7, 0x9f, 0xe3, 0x9f, 0xe6, 0xb0, 0x44, 0x21, 0x06, 0x30, 0x7e,
    0x06, 0x03, 0x55, 0x1d, 0x11, 0x04, 0x77, 0x30, 0x75, 0x82, 0x10, 0x77,
    0x77, 0x77, 0x2e, 0x64, 0x69, 0x67, 0x69, 0x63, 0x65, 0x72, 0x74, 0x2e,
    0x63, 0x6f, 0x6d, 0x82, 0x0c, 0x64, 0x69, 0x67, 0x69, 0x63, 0x65, 0x72,
    0x74, 0x2e, 0x63, 0x6f, 0x6d, 0x82, 0x14, 0x63, 0x6f, 0x6e, 0x74, 0x65,
    0x6e, 0x74, 0x2e, 0x64, 0x69, 0x67, 0x69, 0x63, 0x65, 0x72, 0x74, 0x2e,
    0x63, 0x6f, 0x6d, 0x82, 0x17, 0x77, 0x77, 0x77, 0x2e, 0x6f, 0x72, 0x69,
    0x67, 0x69, 0x6e, 0x2e, 0x64, 0x69, 0x67, 0x69, 0x63, 0x65, 0x72, 0x74,
    0x2e, 0x63, 0x6f, 0x6d, 0x82, 0x12, 0x6c, 0x6f, 0x67, 0x69, 0x6e, 0x2e,
    0x64, 0x69, 0x67, 0x69, 0x63, 0x65, 0x72, 0x74, 0x2e, 0x63, 0x6f, 0x6d,
    0x82, 0x10, 0x61, 0x70, 0x69, 0x2e, 0x64, 0x69, 0x67, 0x69, 0x63, 0x65,
    0x72, 0x74, 0x2e, 0x63, 0x6f, 0x6d, 0x30, 0x0e, 0x06, 0x03, 0x55, 0x1d,
    0x0f, 0x01, 0x01, 0xff, 0x04, 0x04, 0x03, 0x02, 0x05, 0xa0, 0x30, 0x1d,
    0x06, 0x03, 0x55, 0x1d, 0x25, 0x04, 0x16, 0x30, 0x14, 0x06, 0x08, 0x2b,
    0x06, 0x01, 0x05, 0x05, 0x07, 0x03, 0x01, 0x06, 0x08, 0x2b, 0x06, 0x01,
    0x05, 0x05, 0x07, 0x03, 0x02, 0x30, 0x75, 0x06, 0x03, 0x55, 0x1d, 0x1f,
    0x04, 0x6e, 0x30, 0x6c, 0x30, 0x34, 0xa0, 0x32, 0xa0, 0x30, 0x86, 0x2e,
    0x68, 0x74, 0x74, 0x70, 0x3a, 0x2f, 0x2f, 0x63, 0x72, 0x6c, 0x33, 0x2e,
    0x64, 0x69, 0x67, 0x69, 0x63, 0x65, 0x72, 0x74, 0x2e, 0x63, 0x6f, 0x6d,
    0x2f, 0x73, 0x68, 0x61, 0x32, 0x2d, 0x65, 0x76, 0x2d, 0x73, 0x65, 0x72,
    0x76, 0x65, 0x72, 0x2d, 0x67, 0x31, 0x2e, 0x63, 0x72, 0x6c, 0x30, 0x34,
    0xa0, 0x32, 0xa0, 0x30, 0x86, 0x2e, 0x68, 0x74, 0x74, 0x70, 0x3a, 0x2f,
    0x2f, 0x63, 0x72, 0x6c, 0x34, 0x2e, 0x64, 0x69, 0x67, 0x69, 0x63, 0x65,
    0x72, 0x74, 0x2e, 0x63, 0x6f, 0x6d, 0x2f, 0x73, 0x68, 0x61, 0x32, 0x2d,
    0x65, 0x76, 0x2d, 0x73, 0x65, 0x72, 0x76, 0x65, 0x72, 0x2d, 0x67, 0x31,
    0x2e, 0x63, 0x72, 0x6c, 0x30, 0x42, 0x06, 0x03, 0x55, 0x1d, 0x20, 0x04,
    0x3b, 0x30, 0x39, 0x30, 0x37, 0x06, 0x09, 0x60, 0x86, 0x48, 0x01, 0x86,
    0xfd, 0x6c, 0x02, 0x01, 0x30, 0x2a, 0x30, 0x28, 0x06, 0x08, 0x2b, 0x06,
    0x01, 0x05, 0x05, 0x07, 0x02, 0x01, 0x16, 0x1c, 0x68, 0x74, 0x74, 0x70,
    0x73, 0x3a, 0x2f, 0x2f, 0x77, 0x77, 0x77, 0x2e, 0x64, 0x69, 0x67, 0x69,
    0x63, 0x65, 0x72, 0x74, 0x2e, 0x63, 0x6f, 0x6d, 0x2f, 0x43, 0x50, 0x53,
    0x30, 0x81, 0x88, 0x06, 0x08, 0x2b, 0x06, 0x01, 0x05, 0x05, 0x07, 0x01,
    0x01, 0x04, 0x7c, 0x30, 0x7a, 0x30, 0x24, 0x06, 0x08, 0x2b, 0x06, 0x01,
    0x05, 0x05, 0x07, 0x30, 0x01, 0x86, 0x18, 0x68, 0x74, 0x74, 0x70, 0x3a,
    0x2f, 0x2f, 0x6f, 0x63, 0x73, 0x70, 0x2e, 0x64, 0x69, 0x67, 0x69, 0x63,
    0x65, 0x72, 0x74, 0x2e, 0x63, 0x6f, 0x6d, 0x30, 0x52, 0x06, 0x08, 0x2b,
    0x06, 0x01, 0x05, 0x05, 0x07, 0x30, 0x02, 0x86, 0x46, 0x68, 0x74, 0x74,
    0x70, 0x3a, 0x2f, 0x2f, 0x63, 0x61, 0x63, 0x65, 0x72, 0x74, 0x73, 0x2e,
    0x64, 0x69, 0x67, 0x69, 0x63, 0x65, 0x72, 0x74, 0x2e, 0x63, 0x6f, 0x6d,
    0x2f, 0x44, 0x69, 0x67, 0x69, 0x43, 0x65, 0x72, 0x74, 0x53, 0x48, 0x41,
    0x32, 0x45, 0x78, 0x74, 0x65, 0x6e, 0x64, 0x65, 0x64, 0x56, 0x61, 0x6c,
    0x69, 0x64, 0x61, 0x74, 0x69, 0x6f, 0x6e, 0x53, 0x65, 0x72, 0x76, 0x65,
    0x72, 0x43, 0x41, 0x2e, 0x63, 0x72, 0x74, 0x30, 0x0c, 0x06, 0x03, 0x55,
    0x1d, 0x13, 0x01, 0x01, 0xff, 0x04, 0x02, 0x30, 0x00, 0x30, 0x82, 0x01,
    0x03, 0x06, 0x0a, 0x2b, 0x06, 0x01, 0x04, 0x01, 0xd6, 0x79, 0x02, 0x04,
    0x02, 0x04, 0x81, 0xf4, 0x04, 0x81, 0xf1, 0x00, 0xef, 0x00, 0x76, 0x00,
    0xa4, 0xb9, 0x09, 0x90, 0xb4, 0x18, 0x58, 0x14, 0x87, 0xbb, 0x13, 0xa2,
    0xcc, 0x67, 0x70, 0x0a, 0x3c, 0x35, 0x98, 0x04, 0xf9, 0x1b, 0xdf, 0xb8,
    0xe3, 0x77, 0xcd, 0x0e, 0xc8, 0x0d, 0xdc, 0x10, 0x00, 0x00, 0x01, 0x49,
    0x10, 0xfa, 0xbd, 0x89, 0x00, 0x00, 0x04, 0x03, 0x00, 0x47, 0x30, 0x45,
    0x02, 0x20, 0x66, 0xd7, 0x67, 0x79, 0xf4, 0xaa, 0xd3, 0xb8, 0xc6, 0x9f,
    0x03, 0x01, 0xbf, 0xcd, 0xec, 0x83, 0x36, 0xd4, 0xc8, 0x4f, 0xc1, 0x45,
    0xd5, 0xd9, 0xfd, 0x16, 0x54, 0xad, 0x6f, 0x75, 0x22, 0xa1, 0x02, 0x21,
    0x00, 0xb8, 0x95, 0xf1, 0x43, 0x03, 0xdf, 0xa4, 0x11, 0x04, 0x3c, 0x24,
    0x13, 0xd8, 0x81, 0x69, 0x24, 0x9d, 0xd2, 0x04, 0x96, 0x4d, 0xad, 0x53,
    0x3d, 0x9d, 0x6a, 0x24, 0x14, 0x32, 0x4d, 0xcc, 0x91, 0x00, 0x75, 0x00,
    0x68, 0xf6, 0x98, 0xf8, 0x1f, 0x64, 0x82, 0xbe, 0x3a, 0x8c, 0xee, 0xb9,
    0x28, 0x1d, 0x4c, 0xfc, 0x71, 0x51, 0x5d, 0x67, 0x93, 0xd4, 0x44, 0xd1,
    0x0a, 0x67, 0xac, 0xbb, 0x4f, 0x4f, 0xfb, 0xc4, 0x00, 0x00, 0x01, 0x49,
    0x10, 0xfa, 0xbd, 0x79, 0x00, 0x00, 0x04, 0x03, 0x00, 0x46, 0x30, 0x44,
    0x02, 0x20, 0x11, 0x34, 0x9a, 0x59, 0x2c, 0x9d, 0x3b, 0xd3, 0x8b, 0x9a,
    0x58, 0x18, 0x37, 0x24, 0x55, 0xf3, 0x9d, 0x0e, 0xca, 0x98, 0x96, 0x6b,
    0x8f, 0xc7, 0xa2, 0xe4, 0xd8, 0xbf, 0x00, 0xce, 0x40, 0xfd, 0x02, 0x20,
    0x11, 0x24, 0x11, 0xab, 0x62, 0x7f, 0xb2, 0x88, 0xf0, 0x6d, 0x70, 0xc0,
    0xfd, 0xa0, 0x65, 0xb5, 0xb6, 0x03, 0x46, 0x1f, 0x10, 0x30, 0xed, 0xf5,
    0x6d, 0x7e, 0x89, 0x7b, 0xba, 0x20, 0x32, 0x64, 0x30, 0x0d, 0x06, 0x09,
    0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01, 0x0b, 0x05, 0x00, 0x03,
    0x82, 0x01, 0x01, 0x00, 0x7a, 0x12, 0xd4, 0x28, 0xc3, 0x27, 0x39, 0x5e,
    0xab, 0x30, 0x4a, 0x4a, 0x3d, 0xe6, 0xad, 0x78, 0xd7, 0xe7, 0x9d, 0x5a,
    0xcb, 0x93, 0xcf, 0xc2, 0x30, 0x91, 0xfe, 0xcd, 0x88, 0x65, 0x35, 0x96,
    0x9c, 0x5e, 0x6d, 0xe1, 0xe2, 0x6d, 0x00, 0x7e, 0xe2, 0x73, 0xab, 0xd8,
    0x27, 0x97, 0x6b, 0xdd, 0x46, 0xa9, 0x08, 0x41, 0x1e, 0x0b, 0x18, 0x5b,
    0x6a, 0x5b, 0xf8, 0xfc, 0xa4, 0x75, 0xf5, 0x5e, 0xf5, 0x52, 0xec, 0xa5,
    0xab, 0x6b, 0x16, 0x23, 0x44, 0xdf, 0xe8, 0x14, 0x28, 0xe7, 0xd1, 0xbf,
    0xcc, 0x2d, 0x72, 0xa2, 0x30, 0xc4, 0x9e, 0x9a, 0xe5, 0x8c, 0x21, 0xb6,
    0x5c, 0x6f, 0x9e, 0x16, 0xfc, 0xf2, 0x83, 0xe6, 0xf7, 0x29, 0x54, 0xa7,
    0x58, 0x8e, 0x86, 0xb0, 0x8e, 0x5b, 0xf3, 0xd4, 0x1a, 0xaf, 0x00, 0x64,
    0x37, 0xca, 0x2f, 0x0d, 0x38, 0x34, 0x16, 0x96, 0x57, 0x7d, 0x55, 0x6d,
    0x09, 0x2f, 0x57, 0x47, 0x27, 0x0c, 0xd3, 0x02, 0x0e, 0x60, 0x66, 0x91,
    0x92, 0x47, 0x11, 0xac, 0x04, 0xda, 0x63, 0x86, 0x91, 0x32, 0xd1, 0xfa,
    0x65, 0x2c, 0xb7, 0x76, 0x64, 0x2a, 0x0f, 0x25, 0x83, 0x17, 0x3f, 0x0e,
    0x88, 0xf7, 0xeb, 0xf5, 0xf5, 0x68, 0x30, 0x68, 0xcc, 0x79, 0x32, 0x23,
    0x03, 0x08, 0xec, 0x30, 0x1b, 0x05, 0x37, 0x14, 0x79, 0x93, 0xf0, 0xed,
    0x62, 0x88, 0xd8, 0xab, 0x18, 0xc4, 0x98, 0x67, 0xcb, 0x32, 0xbe, 0x44,
    0xa8, 0x8f, 0xbc, 0xff, 0xe7, 0x99, 0x07, 0xc3, 0xd0, 0x4b, 0x0d, 0x45,
    0xf2, 0xa6, 0xd8, 0x84, 0x43, 0xeb, 0xf4, 0x96, 0x3c, 0x3c, 0x68, 0x8c,
    0x81, 0x4f, 0x82, 0x1e, 0x9d, 0xdb, 0x78, 0xa0, 0x42, 0x54, 0x94, 0xf2,
    0x93, 0x5f, 0x49, 0x21, 0x57, 0x1b, 0xe5, 0x86, 0x2b, 0x54, 0xaa, 0xb3,
    0x38, 0x48, 0x61, 0x3e, 0x85, 0x0f, 0x7d, 0x50
};


static
unsigned char digicert_sha2_ev_server_ca_der [] = {
    0x30, 0x82, 0x04, 0xb6, 0x30, 0x82, 0x03, 0x9e, 0xa0, 0x03, 0x02, 0x01,
    0x02, 0x02, 0x10, 0x0c, 0x79, 0xa9, 0x44, 0xb0, 0x8c, 0x11, 0x95, 0x20,
    0x92, 0x61, 0x5f, 0xe2, 0x6b, 0x1d, 0x83, 0x30, 0x0d, 0x06, 0x09, 0x2a,
    0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01, 0x0b, 0x05, 0x00, 0x30, 0x6c,
    0x31, 0x0b, 0x30, 0x09, 0x06, 0x03, 0x55, 0x04, 0x06, 0x13, 0x02, 0x55,
    0x53, 0x31, 0x15, 0x30, 0x13, 0x06, 0x03, 0x55, 0x04, 0x0a, 0x13, 0x0c,
    0x44, 0x69, 0x67, 0x69, 0x43, 0x65, 0x72, 0x74, 0x20, 0x49, 0x6e, 0x63,
    0x31, 0x19, 0x30, 0x17, 0x06, 0x03, 0x55, 0x04, 0x0b, 0x13, 0x10, 0x77,
    0x77, 0x77, 0x2e, 0x64, 0x69, 0x67, 0x69, 0x63, 0x65, 0x72, 0x74, 0x2e,
    0x63, 0x6f, 0x6d, 0x31, 0x2b, 0x30, 0x29, 0x06, 0x03, 0x55, 0x04, 0x03,
    0x13, 0x22, 0x44, 0x69, 0x67, 0x69, 0x43, 0x65, 0x72, 0x74, 0x20, 0x48,
    0x69, 0x67, 0x68, 0x20, 0x41, 0x73, 0x73, 0x75, 0x72, 0x61, 0x6e, 0x63,
    0x65, 0x20, 0x45, 0x56, 0x20, 0x52, 0x6f, 0x6f, 0x74, 0x20, 0x43, 0x41,
    0x30, 0x1e, 0x17, 0x0d, 0x31, 0x33, 0x31, 0x30, 0x32, 0x32, 0x31, 0x32,
    0x30, 0x30, 0x30, 0x30, 0x5a, 0x17, 0x0d, 0x32, 0x38, 0x31, 0x30, 0x32,
    0x32, 0x31, 0x32, 0x30, 0x30, 0x30, 0x30, 0x5a, 0x30, 0x75, 0x31, 0x0b,
    0x30, 0x09, 0x06, 0x03, 0x55, 0x04, 0x06, 0x13, 0x02, 0x55, 0x53, 0x31,
    0x15, 0x30, 0x13, 0x06, 0x03, 0x55, 0x04, 0x0a, 0x13, 0x0c, 0x44, 0x69,
    0x67, 0x69, 0x43, 0x65, 0x72, 0x74, 0x20, 0x49, 0x6e, 0x63, 0x31, 0x19,
    0x30, 0x17, 0x06, 0x03, 0x55, 0x04, 0x0b, 0x13, 0x10, 0x77, 0x77, 0x77,
    0x2e, 0x64, 0x69, 0x67, 0x69, 0x63, 0x65, 0x72, 0x74, 0x2e, 0x63, 0x6f,
    0x6d, 0x31, 0x34, 0x30, 0x32, 0x06, 0x03, 0x55, 0x04, 0x03, 0x13, 0x2b,
    0x44, 0x69, 0x67, 0x69, 0x43, 0x65, 0x72, 0x74, 0x20, 0x53, 0x48, 0x41,
    0x32, 0x20, 0x45, 0x78, 0x74, 0x65, 0x6e, 0x64, 0x65, 0x64, 0x20, 0x56,
    0x61, 0x6c, 0x69, 0x64, 0x61, 0x74, 0x69, 0x6f, 0x6e, 0x20, 0x53, 0x65,
    0x72, 0x76, 0x65, 0x72, 0x20, 0x43, 0x41, 0x30, 0x82, 0x01, 0x22, 0x30,
    0x0d, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01, 0x01,
    0x05, 0x00, 0x03, 0x82, 0x01, 0x0f, 0x00, 0x30, 0x82, 0x01, 0x0a, 0x02,
    0x82, 0x01, 0x01, 0x00, 0xd7, 0x53, 0xa4, 0x04, 0x51, 0xf8, 0x99, 0xa6,
    0x16, 0x48, 0x4b, 0x67, 0x27, 0xaa, 0x93, 0x49, 0xd0, 0x39, 0xed, 0x0c,
    0xb0, 0xb0, 0x00, 0x87, 0xf1, 0x67, 0x28, 0x86, 0x85, 0x8c, 0x8e, 0x63,
    0xda, 0xbc, 0xb1, 0x40, 0x38, 0xe2, 0xd3, 0xf5, 0xec, 0xa5, 0x05, 0x18,
    0xb8, 0x3d, 0x3e, 0xc5, 0x99, 0x17, 0x32, 0xec, 0x18, 0x8c, 0xfa, 0xf1,
    0x0c, 0xa6, 0x64, 0x21, 0x85, 0xcb, 0x07, 0x10, 0x34, 0xb0, 0x52, 0x88,
    0x2b, 0x1f, 0x68, 0x9b, 0xd2, 0xb1, 0x8f, 0x12, 0xb0, 0xb3, 0xd2, 0xe7,
    0x88, 0x1f, 0x1f, 0xef, 0x38, 0x77, 0x54, 0x53, 0x5f, 0x80, 0x79, 0x3f,
    0x2e, 0x1a, 0xaa, 0xa8, 0x1e, 0x4b, 0x2b, 0x0d, 0xab, 0xb7, 0x63, 0xb9,
    0x35, 0xb7, 0x7d, 0x14, 0xbc, 0x59, 0x4b, 0xdf, 0x51, 0x4a, 0xd2, 0xa1,
    0xe2, 0x0c, 0xe2, 0x90, 0x82, 0x87, 0x6a, 0xae, 0xea, 0xd7, 0x64, 0xd6,
    0x98, 0x55, 0xe8, 0xfd, 0xaf, 0x1a, 0x50, 0x6c, 0x54, 0xbc, 0x11, 0xf2,
    0xfd, 0x4a, 0xf2, 0x9d, 0xbb, 0x7f, 0x0e, 0xf4, 0xd5, 0xbe, 0x8e, 0x16,
    0x89, 0x12, 0x55, 0xd8, 0xc0, 0x71, 0x34, 0xee, 0xf6, 0xdc, 0x2d, 0xec,
    0xc4, 0x87, 0x25, 0x86, 0x8d, 0xd8, 0x21, 0xe4, 0xb0, 0x4d, 0x0c, 0x89,
    0xdc, 0x39, 0x26, 0x17, 0xdd, 0xf6, 0xd7, 0x94, 0x85, 0xd8, 0x04, 0x21,
    0x70, 0x9d, 0x6f, 0x6f, 0xff, 0x5c, 0xba, 0x19, 0xe1, 0x45, 0xcb, 0x56,
    0x57, 0x28, 0x7e, 0x1c, 0x0d, 0x41, 0x57, 0xaa, 0xb7, 0xb8, 0x27, 0xbb,
    0xb1, 0xe4, 0xfa, 0x2a, 0xef, 0x21, 0x23, 0x75, 0x1a, 0xad, 0x2d, 0x9b,
    0x86, 0x35, 0x8c, 0x9c, 0x77, 0xb5, 0x73, 0xad, 0xd8, 0x94, 0x2d, 0xe4,
    0xf3, 0x0c, 0x9d, 0xee, 0xc1, 0x4e, 0x62, 0x7e, 0x17, 0xc0, 0x71, 0x9e,
    0x2c, 0xde, 0xf1, 0xf9, 0x10, 0x28, 0x19, 0x33, 0x02, 0x03, 0x01, 0x00,
    0x01, 0xa3, 0x82, 0x01, 0x49, 0x30, 0x82, 0x01, 0x45, 0x30, 0x12, 0x06,
    0x03, 0x55, 0x1d, 0x13, 0x01, 0x01, 0xff, 0x04, 0x08, 0x30, 0x06, 0x01,
    0x01, 0xff, 0x02, 0x01, 0x00, 0x30, 0x0e, 0x06, 0x03, 0x55, 0x1d, 0x0f,
    0x01, 0x01, 0xff, 0x04, 0x04, 0x03, 0x02, 0x01, 0x86, 0x30, 0x1d, 0x06,
    0x03, 0x55, 0x1d, 0x25, 0x04, 0x16, 0x30, 0x14, 0x06, 0x08, 0x2b, 0x06,
    0x01, 0x05, 0x05, 0x07, 0x03, 0x01, 0x06, 0x08, 0x2b, 0x06, 0x01, 0x05,
    0x05, 0x07, 0x03, 0x02, 0x30, 0x34, 0x06, 0x08, 0x2b, 0x06, 0x01, 0x05,
    0x05, 0x07, 0x01, 0x01, 0x04, 0x28, 0x30, 0x26, 0x30, 0x24, 0x06, 0x08,
    0x2b, 0x06, 0x01, 0x05, 0x05, 0x07, 0x30, 0x01, 0x86, 0x18, 0x68, 0x74,
    0x74, 0x70, 0x3a, 0x2f, 0x2f, 0x6f, 0x63, 0x73, 0x70, 0x2e, 0x64, 0x69,
    0x67, 0x69, 0x63, 0x65, 0x72, 0x74, 0x2e, 0x63, 0x6f, 0x6d, 0x30, 0x4b,
    0x06, 0x03, 0x55, 0x1d, 0x1f, 0x04, 0x44, 0x30, 0x42, 0x30, 0x40, 0xa0,
    0x3e, 0xa0, 0x3c, 0x86, 0x3a, 0x68, 0x74, 0x74, 0x70, 0x3a, 0x2f, 0x2f,
    0x63, 0x72, 0x6c, 0x34, 0x2e, 0x64, 0x69, 0x67, 0x69, 0x63, 0x65, 0x72,
    0x74, 0x2e, 0x63, 0x6f, 0x6d, 0x2f, 0x44, 0x69, 0x67, 0x69, 0x43, 0x65,
    0x72, 0x74, 0x48, 0x69, 0x67, 0x68, 0x41, 0x73, 0x73, 0x75, 0x72, 0x61,
    0x6e, 0x63, 0x65, 0x45, 0x56, 0x52, 0x6f, 0x6f, 0x74, 0x43, 0x41, 0x2e,
    0x63, 0x72, 0x6c, 0x30, 0x3d, 0x06, 0x03, 0x55, 0x1d, 0x20, 0x04, 0x36,
    0x30, 0x34, 0x30, 0x32, 0x06, 0x04, 0x55, 0x1d, 0x20, 0x00, 0x30, 0x2a,
    0x30, 0x28, 0x06, 0x08, 0x2b, 0x06, 0x01, 0x05, 0x05, 0x07, 0x02, 0x01,
    0x16, 0x1c, 0x68, 0x74, 0x74, 0x70, 0x73, 0x3a, 0x2f, 0x2f, 0x77, 0x77,
    0x77, 0x2e, 0x64, 0x69, 0x67, 0x69, 0x63, 0x65, 0x72, 0x74, 0x2e, 0x63,
    0x6f, 0x6d, 0x2f, 0x43, 0x50, 0x53, 0x30, 0x1d, 0x06, 0x03, 0x55, 0x1d,
    0x0e, 0x04, 0x16, 0x04, 0x14, 0x3d, 0xd3, 0x50, 0xa5, 0xd6, 0xa0, 0xad,
    0xee, 0xf3, 0x4a, 0x60, 0x0a, 0x65, 0xd3, 0x21, 0xd4, 0xf8, 0xf8, 0xd6,
    0x0f, 0x30, 0x1f, 0x06, 0x03, 0x55, 0x1d, 0x23, 0x04, 0x18, 0x30, 0x16,
    0x80, 0x14, 0xb1, 0x3e, 0xc3, 0x69, 0x03, 0xf8, 0xbf, 0x47, 0x01, 0xd4,
    0x98, 0x26, 0x1a, 0x08, 0x02, 0xef, 0x63, 0x64, 0x2b, 0xc3, 0x30, 0x0d,
    0x06, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01, 0x0b, 0x05,
    0x00, 0x03, 0x82, 0x01, 0x01, 0x00, 0x9d, 0xb6, 0xd0, 0x90, 0x86, 0xe1,
    0x86, 0x02, 0xed, 0xc5, 0xa0, 0xf0, 0x34, 0x1c, 0x74, 0xc1, 0x8d, 0x76,
    0xcc, 0x86, 0x0a, 0xa8, 0xf0, 0x4a, 0x8a, 0x42, 0xd6, 0x3f, 0xc8, 0xa9,
    0x4d, 0xad, 0x7c, 0x08, 0xad, 0xe6, 0xb6, 0x50, 0xb8, 0xa2, 0x1a, 0x4d,
    0x88, 0x07, 0xb1, 0x29, 0x21, 0xdc, 0xe7, 0xda, 0xc6, 0x3c, 0x21, 0xe0,
    0xe3, 0x11, 0x49, 0x70, 0xac, 0x7a, 0x1d, 0x01, 0xa4, 0xca, 0x11, 0x3a,
    0x57, 0xab, 0x7d, 0x57, 0x2a, 0x40, 0x74, 0xfd, 0xd3, 0x1d, 0x85, 0x18,
    0x50, 0xdf, 0x57, 0x47, 0x75, 0xa1, 0x7d, 0x55, 0x20, 0x2e, 0x47, 0x37,
    0x50, 0x72, 0x8c, 0x7f, 0x82, 0x1b, 0xd2, 0x62, 0x8f, 0x2d, 0x03, 0x5a,
    0xda, 0xc3, 0xc8, 0xa1, 0xce, 0x2c, 0x52, 0xa2, 0x00, 0x63, 0xeb, 0x73,
    0xba, 0x71, 0xc8, 0x49, 0x27, 0x23, 0x97, 0x64, 0x85, 0x9e, 0x38, 0x0e,
    0xad, 0x63, 0x68, 0x3c, 0xba, 0x52, 0x81, 0x58, 0x79, 0xa3, 0x2c, 0x0c,
    0xdf, 0xde, 0x6d, 0xeb, 0x31, 0xf2, 0xba, 0xa0, 0x7c, 0x6c, 0xf1, 0x2c,
    0xd4, 0xe1, 0xbd, 0x77, 0x84, 0x37, 0x03, 0xce, 0x32, 0xb5, 0xc8, 0x9a,
    0x81, 0x1a, 0x4a, 0x92, 0x4e, 0x3b, 0x46, 0x9a, 0x85, 0xfe, 0x83, 0xa2,
    0xf9, 0x9e, 0x8c, 0xa3, 0xcc, 0x0d, 0x5e, 0xb3, 0x3d, 0xcf, 0x04, 0x78,
    0x8f, 0x14, 0x14, 0x7b, 0x32, 0x9c, 0xc7, 0x00, 0xa6, 0x5c, 0xc4, 0xb5,
    0xa1, 0x55, 0x8d, 0x5a, 0x56, 0x68, 0xa4, 0x22, 0x70, 0xaa, 0x3c, 0x81,
    0x71, 0xd9, 0x9d, 0xa8, 0x45, 0x3b, 0xf4, 0xe5, 0xf6, 0xa2, 0x51, 0xdd,
    0xc7, 0x7b, 0x62, 0xe8, 0x6f, 0x0c, 0x74, 0xeb, 0xb8, 0xda, 0xf8, 0xbf,
    0x87, 0x0d, 0x79, 0x50, 0x91, 0x90, 0x9b, 0x18, 0x3b, 0x91, 0x59, 0x27,
    0xf1, 0x35, 0x28, 0x13, 0xab, 0x26, 0x7e, 0xd5, 0xf7, 0x7a
};


static
unsigned char www_paypal_com_cert_der[] = {
    0x30, 0x82, 0x06, 0x08, 0x30, 0x82, 0x04, 0xf0, 0xa0, 0x03, 0x02, 0x01,
    0x02, 0x02, 0x10, 0x08, 0x34, 0xe4, 0x53, 0xd4, 0x3a, 0x68, 0x57, 0x23,
    0xaf, 0xfb, 0xb1, 0x33, 0xce, 0x45, 0x7c, 0x30, 0x0d, 0x06, 0x09, 0x2a,
    0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01, 0x05, 0x05, 0x00, 0x30, 0x81,
    0xba, 0x31, 0x0b, 0x30, 0x09, 0x06, 0x03, 0x55, 0x04, 0x06, 0x13, 0x02,
    0x55, 0x53, 0x31, 0x17, 0x30, 0x15, 0x06, 0x03, 0x55, 0x04, 0x0a, 0x13,
    0x0e, 0x56, 0x65, 0x72, 0x69, 0x53, 0x69, 0x67, 0x6e, 0x2c, 0x20, 0x49,
    0x6e, 0x63, 0x2e, 0x31, 0x1f, 0x30, 0x1d, 0x06, 0x03, 0x55, 0x04, 0x0b,
    0x13, 0x16, 0x56, 0x65, 0x72, 0x69, 0x53, 0x69, 0x67, 0x6e, 0x20, 0x54,
    0x72, 0x75, 0x73, 0x74, 0x20, 0x4e, 0x65, 0x74, 0x77, 0x6f, 0x72, 0x6b,
    0x31, 0x3b, 0x30, 0x39, 0x06, 0x03, 0x55, 0x04, 0x0b, 0x13, 0x32, 0x54,
    0x65, 0x72, 0x6d, 0x73, 0x20, 0x6f, 0x66, 0x20, 0x75, 0x73, 0x65, 0x20,
    0x61, 0x74, 0x20, 0x68, 0x74, 0x74, 0x70, 0x73, 0x3a, 0x2f, 0x2f, 0x77,
    0x77, 0x77, 0x2e, 0x76, 0x65, 0x72, 0x69, 0x73, 0x69, 0x67, 0x6e, 0x2e,
    0x63, 0x6f, 0x6d, 0x2f, 0x72, 0x70, 0x61, 0x20, 0x28, 0x63, 0x29, 0x30,
    0x36, 0x31, 0x34, 0x30, 0x32, 0x06, 0x03, 0x55, 0x04, 0x03, 0x13, 0x2b,
    0x56, 0x65, 0x72, 0x69, 0x53, 0x69, 0x67, 0x6e, 0x20, 0x43, 0x6c, 0x61,
    0x73, 0x73, 0x20, 0x33, 0x20, 0x45, 0x78, 0x74, 0x65, 0x6e, 0x64, 0x65,
    0x64, 0x20, 0x56, 0x61, 0x6c, 0x69, 0x64, 0x61, 0x74, 0x69, 0x6f, 0x6e,
    0x20, 0x53, 0x53, 0x4c, 0x20, 0x43, 0x41, 0x30, 0x1e, 0x17, 0x0d, 0x31,
    0x34, 0x30, 0x34, 0x31, 0x35, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x5a,
    0x17, 0x0d, 0x31, 0x35, 0x30, 0x34, 0x30, 0x32, 0x32, 0x33, 0x35, 0x39,
    0x35, 0x39, 0x5a, 0x30, 0x82, 0x01, 0x09, 0x31, 0x13, 0x30, 0x11, 0x06,
    0x0b, 0x2b, 0x06, 0x01, 0x04, 0x01, 0x82, 0x37, 0x3c, 0x02, 0x01, 0x03,
    0x13, 0x02, 0x55, 0x53, 0x31, 0x19, 0x30, 0x17, 0x06, 0x0b, 0x2b, 0x06,
    0x01, 0x04, 0x01, 0x82, 0x37, 0x3c, 0x02, 0x01, 0x02, 0x13, 0x08, 0x44,
    0x65, 0x6c, 0x61, 0x77, 0x61, 0x72, 0x65, 0x31, 0x1d, 0x30, 0x1b, 0x06,
    0x03, 0x55, 0x04, 0x0f, 0x13, 0x14, 0x50, 0x72, 0x69, 0x76, 0x61, 0x74,
    0x65, 0x20, 0x4f, 0x72, 0x67, 0x61, 0x6e, 0x69, 0x7a, 0x61, 0x74, 0x69,
    0x6f, 0x6e, 0x31, 0x10, 0x30, 0x0e, 0x06, 0x03, 0x55, 0x04, 0x05, 0x13,
    0x07, 0x33, 0x30, 0x31, 0x34, 0x32, 0x36, 0x37, 0x31, 0x0b, 0x30, 0x09,
    0x06, 0x03, 0x55, 0x04, 0x06, 0x13, 0x02, 0x55, 0x53, 0x31, 0x13, 0x30,
    0x11, 0x06, 0x03, 0x55, 0x04, 0x11, 0x14, 0x0a, 0x39, 0x35, 0x31, 0x33,
    0x31, 0x2d, 0x32, 0x30, 0x32, 0x31, 0x31, 0x13, 0x30, 0x11, 0x06, 0x03,
    0x55, 0x04, 0x08, 0x13, 0x0a, 0x43, 0x61, 0x6c, 0x69, 0x66, 0x6f, 0x72,
    0x6e, 0x69, 0x61, 0x31, 0x11, 0x30, 0x0f, 0x06, 0x03, 0x55, 0x04, 0x07,
    0x14, 0x08, 0x53, 0x61, 0x6e, 0x20, 0x4a, 0x6f, 0x73, 0x65, 0x31, 0x16,
    0x30, 0x14, 0x06, 0x03, 0x55, 0x04, 0x09, 0x14, 0x0d, 0x32, 0x32, 0x31,
    0x31, 0x20, 0x4e, 0x20, 0x31, 0x73, 0x74, 0x20, 0x53, 0x74, 0x31, 0x15,
    0x30, 0x13, 0x06, 0x03, 0x55, 0x04, 0x0a, 0x14, 0x0c, 0x50, 0x61, 0x79,
    0x50, 0x61, 0x6c, 0x2c, 0x20, 0x49, 0x6e, 0x63, 0x2e, 0x31, 0x14, 0x30,
    0x12, 0x06, 0x03, 0x55, 0x04, 0x0b, 0x14, 0x0b, 0x43, 0x44, 0x4e, 0x20,
    0x53, 0x75, 0x70, 0x70, 0x6f, 0x72, 0x74, 0x31, 0x17, 0x30, 0x15, 0x06,
    0x03, 0x55, 0x04, 0x03, 0x14, 0x0e, 0x77, 0x77, 0x77, 0x2e, 0x70, 0x61,
    0x79, 0x70, 0x61, 0x6c, 0x2e, 0x63, 0x6f, 0x6d, 0x30, 0x82, 0x01, 0x22,
    0x30, 0x0d, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01,
    0x01, 0x05, 0x00, 0x03, 0x82, 0x01, 0x0f, 0x00, 0x30, 0x82, 0x01, 0x0a,
    0x02, 0x82, 0x01, 0x01, 0x00, 0xbe, 0xae, 0x46, 0x4d, 0x99, 0x6e, 0x6d,
    0x6c, 0x35, 0x4b, 0x88, 0x32, 0x38, 0xbb, 0xdc, 0xd0, 0x09, 0x95, 0xd0,
    0x9a, 0xe4, 0x36, 0xe7, 0x9f, 0x0a, 0xb0, 0xf2, 0xd7, 0xd2, 0x30, 0x62,
    0x03, 0x1f, 0xad, 0xc6, 0xf4, 0x6d, 0x10, 0x84, 0xf7, 0x79, 0x1b, 0xbc,
    0x74, 0xc0, 0xa8, 0xe3, 0x82, 0xfe, 0xd4, 0x0a, 0x93, 0x2e, 0x3d, 0x4b,
    0x12, 0x24, 0xad, 0xad, 0x5f, 0x5d, 0xed, 0x1c, 0xc9, 0x1c, 0x6f, 0x13,
    0x7b, 0xe2, 0xc1, 0x25, 0x4e, 0x46, 0x5f, 0x4f, 0x3b, 0x2e, 0x5a, 0xcb,
    0xc1, 0x5a, 0xb4, 0x82, 0xcf, 0xad, 0xa3, 0x65, 0xe8, 0x86, 0x33, 0xb5,
    0xed, 0x1d, 0x78, 0x99, 0xa7, 0xc7, 0xd5, 0xfa, 0x10, 0x2e, 0xfb, 0x11,
    0x4e, 0x23, 0x58, 0x06, 0x96, 0x87, 0x71, 0x75, 0x51, 0x73, 0x8c, 0x0f,
    0xf4, 0xca, 0x7c, 0x8f, 0x91, 0x25, 0x79, 0x13, 0xdc, 0xb0, 0xf0, 0xde,
    0x08, 0x07, 0x01, 0x0b, 0x64, 0xcc, 0x57, 0x6a, 0x12, 0x86, 0x62, 0x17,
    0x3e, 0x5d, 0xb9, 0x62, 0x3d, 0x58, 0x7b, 0x2a, 0x6e, 0xf6, 0xa6, 0x30,
    0x41, 0x02, 0xfc, 0xec, 0x64, 0x72, 0x33, 0xd5, 0xd5, 0x3f, 0x6b, 0x6d,
    0x97, 0xf3, 0xc1, 0x61, 0xbf, 0x38, 0x3b, 0xab, 0x41, 0x47, 0xd4, 0xc2,
    0x03, 0xd7, 0x3b, 0x59, 0x57, 0x9d, 0xe1, 0xa1, 0x2a, 0xd6, 0x78, 0xe8,
    0x83, 0x5d, 0x3d, 0xdd, 0xaa, 0x5d, 0x17, 0xfd, 0x94, 0xd6, 0xe5, 0x7a,
    0xef, 0x02, 0x63, 0xc6, 0xa3, 0xc6, 0x2d, 0x5b, 0x33, 0x08, 0x8b, 0xf5,
    0xa5, 0x03, 0xb4, 0xfe, 0xf2, 0x1d, 0xab, 0xbf, 0x5e, 0x9e, 0xb8, 0x78,
    0x39, 0x20, 0x2b, 0x68, 0x61, 0x4f, 0xe4, 0x99, 0xf2, 0xaa, 0xc2, 0x4d,
    0x4b, 0x48, 0xcb, 0x68, 0xc2, 0x10, 0x3f, 0xfa, 0x9a, 0xba, 0xc5, 0x6a,
    0x53, 0x8f, 0x22, 0xf3, 0xd7, 0xc9, 0xed, 0xa4, 0xd5, 0x02, 0x03, 0x01,
    0x00, 0x01, 0xa3, 0x82, 0x01, 0xb6, 0x30, 0x82, 0x01, 0xb2, 0x30, 0x67,
    0x06, 0x03, 0x55, 0x1d, 0x11, 0x04, 0x60, 0x30, 0x5e, 0x82, 0x0e, 0x77,
    0x77, 0x77, 0x2e, 0x70, 0x61, 0x79, 0x70, 0x61, 0x6c, 0x2e, 0x63, 0x6f,
    0x6d, 0x82, 0x12, 0x68, 0x69, 0x73, 0x74, 0x6f, 0x72, 0x79, 0x2e, 0x70,
    0x61, 0x79, 0x70, 0x61, 0x6c, 0x2e, 0x63, 0x6f, 0x6d, 0x82, 0x0c, 0x74,
    0x2e, 0x70, 0x61, 0x79, 0x70, 0x61, 0x6c, 0x2e, 0x63, 0x6f, 0x6d, 0x82,
    0x0c, 0x63, 0x2e, 0x70, 0x61, 0x79, 0x70, 0x61, 0x6c, 0x2e, 0x63, 0x6f,
    0x6d, 0x82, 0x0e, 0x74, 0x6d, 0x73, 0x2e, 0x70, 0x61, 0x79, 0x70, 0x61,
    0x6c, 0x2e, 0x63, 0x6f, 0x6d, 0x82, 0x0c, 0x74, 0x6d, 0x73, 0x2e, 0x65,
    0x62, 0x61, 0x79, 0x2e, 0x63, 0x6f, 0x6d, 0x30, 0x09, 0x06, 0x03, 0x55,
    0x1d, 0x13, 0x04, 0x02, 0x30, 0x00, 0x30, 0x0e, 0x06, 0x03, 0x55, 0x1d,
    0x0f, 0x01, 0x01, 0xff, 0x04, 0x04, 0x03, 0x02, 0x05, 0xa0, 0x30, 0x1d,
    0x06, 0x03, 0x55, 0x1d, 0x25, 0x04, 0x16, 0x30, 0x14, 0x06, 0x08, 0x2b,
    0x06, 0x01, 0x05, 0x05, 0x07, 0x03, 0x01, 0x06, 0x08, 0x2b, 0x06, 0x01,
    0x05, 0x05, 0x07, 0x03, 0x02, 0x30, 0x66, 0x06, 0x03, 0x55, 0x1d, 0x20,
    0x04, 0x5f, 0x30, 0x5d, 0x30, 0x5b, 0x06, 0x0b, 0x60, 0x86, 0x48, 0x01,
    0x86, 0xf8, 0x45, 0x01, 0x07, 0x17, 0x06, 0x30, 0x4c, 0x30, 0x23, 0x06,
    0x08, 0x2b, 0x06, 0x01, 0x05, 0x05, 0x07, 0x02, 0x01, 0x16, 0x17, 0x68,
    0x74, 0x74, 0x70, 0x73, 0x3a, 0x2f, 0x2f, 0x64, 0x2e, 0x73, 0x79, 0x6d,
    0x63, 0x62, 0x2e, 0x63, 0x6f, 0x6d, 0x2f, 0x63, 0x70, 0x73, 0x30, 0x25,
    0x06, 0x08, 0x2b, 0x06, 0x01, 0x05, 0x05, 0x07, 0x02, 0x02, 0x30, 0x19,
    0x1a, 0x17, 0x68, 0x74, 0x74, 0x70, 0x73, 0x3a, 0x2f, 0x2f, 0x64, 0x2e,
    0x73, 0x79, 0x6d, 0x63, 0x62, 0x2e, 0x63, 0x6f, 0x6d, 0x2f, 0x72, 0x70,
    0x61, 0x30, 0x1f, 0x06, 0x03, 0x55, 0x1d, 0x23, 0x04, 0x18, 0x30, 0x16,
    0x80, 0x14, 0xfc, 0x8a, 0x50, 0xba, 0x9e, 0xb9, 0x25, 0x5a, 0x7b, 0x55,
    0x85, 0x4f, 0x95, 0x00, 0x63, 0x8f, 0xe9, 0x58, 0x6b, 0x43, 0x30, 0x2b,
    0x06, 0x03, 0x55, 0x1d, 0x1f, 0x04, 0x24, 0x30, 0x22, 0x30, 0x20, 0xa0,
    0x1e, 0xa0, 0x1c, 0x86, 0x1a, 0x68, 0x74, 0x74, 0x70, 0x3a, 0x2f, 0x2f,
    0x73, 0x61, 0x2e, 0x73, 0x79, 0x6d, 0x63, 0x62, 0x2e, 0x63, 0x6f, 0x6d,
    0x2f, 0x73, 0x61, 0x2e, 0x63, 0x72, 0x6c, 0x30, 0x57, 0x06, 0x08, 0x2b,
    0x06, 0x01, 0x05, 0x05, 0x07, 0x01, 0x01, 0x04, 0x4b, 0x30, 0x49, 0x30,
    0x1f, 0x06, 0x08, 0x2b, 0x06, 0x01, 0x05, 0x05, 0x07, 0x30, 0x01, 0x86,
    0x13, 0x68, 0x74, 0x74, 0x70, 0x3a, 0x2f, 0x2f, 0x73, 0x61, 0x2e, 0x73,
    0x79, 0x6d, 0x63, 0x64, 0x2e, 0x63, 0x6f, 0x6d, 0x30, 0x26, 0x06, 0x08,
    0x2b, 0x06, 0x01, 0x05, 0x05, 0x07, 0x30, 0x02, 0x86, 0x1a, 0x68, 0x74,
    0x74, 0x70, 0x3a, 0x2f, 0x2f, 0x73, 0x61, 0x2e, 0x73, 0x79, 0x6d, 0x63,
    0x62, 0x2e, 0x63, 0x6f, 0x6d, 0x2f, 0x73, 0x61, 0x2e, 0x63, 0x72, 0x74,
    0x30, 0x0d, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01,
    0x05, 0x05, 0x00, 0x03, 0x82, 0x01, 0x01, 0x00, 0x76, 0x08, 0xab, 0x64,
    0xf6, 0xf4, 0x0b, 0xe4, 0x81, 0xbd, 0x59, 0xb2, 0x3e, 0xa4, 0xfc, 0xf5,
    0x03, 0x75, 0x04, 0x59, 0x6a, 0xb5, 0xfe, 0x12, 0x34, 0x2a, 0x04, 0x9c,
    0x89, 0xcd, 0xcb, 0xe1, 0x3c, 0x6c, 0x20, 0x39, 0xd4, 0xea, 0x6f, 0x27,
    0x34, 0x7f, 0x62, 0x1c, 0x45, 0x72, 0x11, 0x39, 0xc0, 0x45, 0xaa, 0x2a,
    0x35, 0x5c, 0xb6, 0x06, 0xe3, 0x08, 0xa7, 0x8f, 0x08, 0xaf, 0x80, 0xb2,
    0x10, 0xce, 0xa5, 0x28, 0x5b, 0x1c, 0x49, 0x55, 0x11, 0xeb, 0x6b, 0x2a,
    0x80, 0xc1, 0x09, 0xed, 0x82, 0x72, 0x48, 0xca, 0x19, 0x8b, 0xe5, 0x34,
    0x94, 0x3c, 0x50, 0x26, 0x77, 0x6b, 0x1a, 0x63, 0xba, 0x6f, 0x63, 0xd1,
    0x58, 0xed, 0x2b, 0x1d, 0xb7, 0xa7, 0x6e, 0x04, 0x25, 0x99, 0xc3, 0x94,
    0x03, 0x90, 0xec, 0x0f, 0x4c, 0x93, 0x83, 0x35, 0x86, 0xe3, 0x70, 0x84,
    0x0d, 0x3c, 0xce, 0xaf, 0x4e, 0x80, 0x4a, 0xd3, 0x91, 0x3f, 0x55, 0x33,
    0x2f, 0x1f, 0x67, 0x87, 0x2f, 0x09, 0xa2, 0x41, 0xc0, 0x10, 0x4a, 0x2c,
    0xc4, 0x88, 0xa0, 0x6f, 0x93, 0x2c, 0xef, 0x38, 0xd2, 0x61, 0xc7, 0xec,
    0xf3, 0x37, 0x7d, 0xc9, 0x32, 0xa5, 0x5c, 0x1e, 0x48, 0x0e, 0x85, 0x6c,
    0x47, 0x2a, 0x7f, 0xc6, 0x30, 0x5e, 0xc2, 0xf6, 0x2e, 0xdd, 0xe3, 0x4d,
    0xac, 0xff, 0xef, 0x48, 0x26, 0xc7, 0x51, 0x74, 0x47, 0x32, 0x46, 0x0b,
    0xcd, 0x7a, 0x0a, 0x5d, 0x5b, 0xc5, 0x8d, 0xed, 0x17, 0xbc, 0xde, 0x09,
    0xbc, 0xe9, 0x93, 0xa9, 0x7c, 0x85, 0x9c, 0x88, 0xa6, 0x83, 0xbc, 0xd6,
    0xe5, 0x1f, 0x05, 0x10, 0xdf, 0xb2, 0x4f, 0xa2, 0xc5, 0x97, 0x00, 0x8b,
    0x57, 0xc7, 0x0d, 0xe7, 0xc7, 0x57, 0x57, 0x87, 0x7d, 0x13, 0x9f, 0x5c,
    0x5c, 0xf7, 0xf3, 0xcd, 0x00, 0x89, 0x0d, 0x85, 0x9a, 0xa2, 0x70, 0xda
};


static
unsigned char www_paypal_com_issuer_cert_der[] = {
    0x30, 0x82, 0x05, 0xe4, 0x30, 0x82, 0x04, 0xcc, 0xa0, 0x03, 0x02, 0x01,
    0x02, 0x02, 0x10, 0x5b, 0x77, 0x59, 0xc6, 0x17, 0x84, 0xe1, 0x5e, 0xc7,
    0x27, 0xc0, 0x32, 0x95, 0x29, 0x28, 0x6b, 0x30, 0x0d, 0x06, 0x09, 0x2a,
    0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01, 0x05, 0x05, 0x00, 0x30, 0x81,
    0xca, 0x31, 0x0b, 0x30, 0x09, 0x06, 0x03, 0x55, 0x04, 0x06, 0x13, 0x02,
    0x55, 0x53, 0x31, 0x17, 0x30, 0x15, 0x06, 0x03, 0x55, 0x04, 0x0a, 0x13,
    0x0e, 0x56, 0x65, 0x72, 0x69, 0x53, 0x69, 0x67, 0x6e, 0x2c, 0x20, 0x49,
    0x6e, 0x63, 0x2e, 0x31, 0x1f, 0x30, 0x1d, 0x06, 0x03, 0x55, 0x04, 0x0b,
    0x13, 0x16, 0x56, 0x65, 0x72, 0x69, 0x53, 0x69, 0x67, 0x6e, 0x20, 0x54,
    0x72, 0x75, 0x73, 0x74, 0x20, 0x4e, 0x65, 0x74, 0x77, 0x6f, 0x72, 0x6b,
    0x31, 0x3a, 0x30, 0x38, 0x06, 0x03, 0x55, 0x04, 0x0b, 0x13, 0x31, 0x28,
    0x63, 0x29, 0x20, 0x32, 0x30, 0x30, 0x36, 0x20, 0x56, 0x65, 0x72, 0x69,
    0x53, 0x69, 0x67, 0x6e, 0x2c, 0x20, 0x49, 0x6e, 0x63, 0x2e, 0x20, 0x2d,
    0x20, 0x46, 0x6f, 0x72, 0x20, 0x61, 0x75, 0x74, 0x68, 0x6f, 0x72, 0x69,
    0x7a, 0x65, 0x64, 0x20, 0x75, 0x73, 0x65, 0x20, 0x6f, 0x6e, 0x6c, 0x79,
    0x31, 0x45, 0x30, 0x43, 0x06, 0x03, 0x55, 0x04, 0x03, 0x13, 0x3c, 0x56,
    0x65, 0x72, 0x69, 0x53, 0x69, 0x67, 0x6e, 0x20, 0x43, 0x6c, 0x61, 0x73,
    0x73, 0x20, 0x33, 0x20, 0x50, 0x75, 0x62, 0x6c, 0x69, 0x63, 0x20, 0x50,
    0x72, 0x69, 0x6d, 0x61, 0x72, 0x79, 0x20, 0x43, 0x65, 0x72, 0x74, 0x69,
    0x66, 0x69, 0x63, 0x61, 0x74, 0x69, 0x6f, 0x6e, 0x20, 0x41, 0x75, 0x74,
    0x68, 0x6f, 0x72, 0x69, 0x74, 0x79, 0x20, 0x2d, 0x20, 0x47, 0x35, 0x30,
    0x1e, 0x17, 0x0d, 0x30, 0x36, 0x31, 0x31, 0x30, 0x38, 0x30, 0x30, 0x30,
    0x30, 0x30, 0x30, 0x5a, 0x17, 0x0d, 0x31, 0x36, 0x31, 0x31, 0x30, 0x37,
    0x32, 0x33, 0x35, 0x39, 0x35, 0x39, 0x5a, 0x30, 0x81, 0xba, 0x31, 0x0b,
    0x30, 0x09, 0x06, 0x03, 0x55, 0x04, 0x06, 0x13, 0x02, 0x55, 0x53, 0x31,
    0x17, 0x30, 0x15, 0x06, 0x03, 0x55, 0x04, 0x0a, 0x13, 0x0e, 0x56, 0x65,
    0x72, 0x69, 0x53, 0x69, 0x67, 0x6e, 0x2c, 0x20, 0x49, 0x6e, 0x63, 0x2e,
    0x31, 0x1f, 0x30, 0x1d, 0x06, 0x03, 0x55, 0x04, 0x0b, 0x13, 0x16, 0x56,
    0x65, 0x72, 0x69, 0x53, 0x69, 0x67, 0x6e, 0x20, 0x54, 0x72, 0x75, 0x73,
    0x74, 0x20, 0x4e, 0x65, 0x74, 0x77, 0x6f, 0x72, 0x6b, 0x31, 0x3b, 0x30,
    0x39, 0x06, 0x03, 0x55, 0x04, 0x0b, 0x13, 0x32, 0x54, 0x65, 0x72, 0x6d,
    0x73, 0x20, 0x6f, 0x66, 0x20, 0x75, 0x73, 0x65, 0x20, 0x61, 0x74, 0x20,
    0x68, 0x74, 0x74, 0x70, 0x73, 0x3a, 0x2f, 0x2f, 0x77, 0x77, 0x77, 0x2e,
    0x76, 0x65, 0x72, 0x69, 0x73, 0x69, 0x67, 0x6e, 0x2e, 0x63, 0x6f, 0x6d,
    0x2f, 0x72, 0x70, 0x61, 0x20, 0x28, 0x63, 0x29, 0x30, 0x36, 0x31, 0x34,
    0x30, 0x32, 0x06, 0x03, 0x55, 0x04, 0x03, 0x13, 0x2b, 0x56, 0x65, 0x72,
    0x69, 0x53, 0x69, 0x67, 0x6e, 0x20, 0x43, 0x6c, 0x61, 0x73, 0x73, 0x20,
    0x33, 0x20, 0x45, 0x78, 0x74, 0x65, 0x6e, 0x64, 0x65, 0x64, 0x20, 0x56,
    0x61, 0x6c, 0x69, 0x64, 0x61, 0x74, 0x69, 0x6f, 0x6e, 0x20, 0x53, 0x53,
    0x4c, 0x20, 0x43, 0x41, 0x30, 0x82, 0x01, 0x22, 0x30, 0x0d, 0x06, 0x09,
    0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01, 0x01, 0x05, 0x00, 0x03,
    0x82, 0x01, 0x0f, 0x00, 0x30, 0x82, 0x01, 0x0a, 0x02, 0x82, 0x01, 0x01,
    0x00, 0x98, 0xdb, 0xa0, 0x55, 0xeb, 0x9c, 0xfd, 0x17, 0x79, 0xe3, 0x9a,
    0x6e, 0x14, 0x1d, 0xb1, 0x5b, 0x98, 0x23, 0x87, 0x16, 0x6e, 0x87, 0x76,
    0x9c, 0xb5, 0x38, 0x3b, 0xb5, 0xa0, 0x7a, 0xb4, 0x07, 0x63, 0x09, 0x19,
    0xe6, 0x2a, 0x88, 0x48, 0xa9, 0xe7, 0x9d, 0xb6, 0x30, 0x5a, 0x08, 0x97,
    0x0c, 0xec, 0xaa, 0xe4, 0x16, 0x69, 0x72, 0x62, 0x23, 0x9a, 0xfb, 0x7a,
    0x54, 0x28, 0x98, 0xc5, 0x0c, 0x2d, 0xb7, 0xd7, 0x22, 0xb6, 0xc8, 0xf9,
    0x38, 0x17, 0xc7, 0xdd, 0xda, 0x31, 0x46, 0x9a, 0x94, 0x14, 0x8e, 0x9e,
    0xee, 0x78, 0xa0, 0xb7, 0x22, 0xd4, 0x49, 0x54, 0x97, 0x4d, 0xe5, 0x74,
    0x5b, 0x92, 0xbc, 0xec, 0x6c, 0x2c, 0xdf, 0xe7, 0xc1, 0xb6, 0x1b, 0x1a,
    0x55, 0x6b, 0x66, 0x08, 0x03, 0x7f, 0x45, 0xaf, 0x9a, 0x33, 0xf1, 0x10,
    0xc0, 0x6c, 0x99, 0x4a, 0x92, 0x24, 0x31, 0x08, 0x6d, 0xdd, 0x02, 0x3e,
    0x61, 0x76, 0x78, 0x78, 0xb6, 0xed, 0x7e, 0x37, 0xae, 0x6c, 0xf3, 0x89,
    0xe1, 0xb7, 0xe1, 0xdc, 0x15, 0xcc, 0xb7, 0x56, 0x9f, 0x80, 0xa0, 0xb1,
    0x05, 0x7f, 0x4e, 0x37, 0x15, 0xff, 0xb7, 0x2f, 0x1e, 0x8f, 0x06, 0x38,
    0x3f, 0x50, 0xb7, 0x69, 0x28, 0xa3, 0xb5, 0x66, 0x5f, 0x36, 0x1a, 0x52,
    0x48, 0x43, 0x66, 0x52, 0xdf, 0xa2, 0x92, 0x4f, 0xd3, 0x18, 0x60, 0xbe,
    0xe3, 0xea, 0x5e, 0x19, 0x71, 0x05, 0xbf, 0x9e, 0x1c, 0x6c, 0x68, 0x72,
    0x25, 0x6f, 0xb3, 0x7b, 0x73, 0xc9, 0x6d, 0xbd, 0x12, 0xff, 0x9b, 0x41,
    0x32, 0x5e, 0xf4, 0xe8, 0x7e, 0xc5, 0x0b, 0xa3, 0x4c, 0x64, 0xd1, 0x4e,
    0xbc, 0x26, 0x08, 0x65, 0xfb, 0x19, 0x97, 0x58, 0x78, 0xe1, 0x33, 0xbf,
    0xed, 0x68, 0x3e, 0xb1, 0x27, 0x45, 0x6f, 0xc0, 0xe2, 0xec, 0x97, 0x69,
    0xf7, 0x5c, 0xd3, 0xf7, 0x51, 0x02, 0x03, 0x01, 0x00, 0x01, 0xa3, 0x82,
    0x01, 0xd2, 0x30, 0x82, 0x01, 0xce, 0x30, 0x1d, 0x06, 0x03, 0x55, 0x1d,
    0x0e, 0x04, 0x16, 0x04, 0x14, 0xfc, 0x8a, 0x50, 0xba, 0x9e, 0xb9, 0x25,
    0x5a, 0x7b, 0x55, 0x85, 0x4f, 0x95, 0x00, 0x63, 0x8f, 0xe9, 0x58, 0x6b,
    0x43, 0x30, 0x12, 0x06, 0x03, 0x55, 0x1d, 0x13, 0x01, 0x01, 0xff, 0x04,
    0x08, 0x30, 0x06, 0x01, 0x01, 0xff, 0x02, 0x01, 0x00, 0x30, 0x3d, 0x06,
    0x03, 0x55, 0x1d, 0x20, 0x04, 0x36, 0x30, 0x34, 0x30, 0x32, 0x06, 0x04,
    0x55, 0x1d, 0x20, 0x00, 0x30, 0x2a, 0x30, 0x28, 0x06, 0x08, 0x2b, 0x06,
    0x01, 0x05, 0x05, 0x07, 0x02, 0x01, 0x16, 0x1c, 0x68, 0x74, 0x74, 0x70,
    0x73, 0x3a, 0x2f, 0x2f, 0x77, 0x77, 0x77, 0x2e, 0x76, 0x65, 0x72, 0x69,
    0x73, 0x69, 0x67, 0x6e, 0x2e, 0x63, 0x6f, 0x6d, 0x2f, 0x63, 0x70, 0x73,
    0x30, 0x3d, 0x06, 0x03, 0x55, 0x1d, 0x1f, 0x04, 0x36, 0x30, 0x34, 0x30,
    0x32, 0xa0, 0x30, 0xa0, 0x2e, 0x86, 0x2c, 0x68, 0x74, 0x74, 0x70, 0x3a,
    0x2f, 0x2f, 0x45, 0x56, 0x53, 0x65, 0x63, 0x75, 0x72, 0x65, 0x2d, 0x63,
    0x72, 0x6c, 0x2e, 0x76, 0x65, 0x72, 0x69, 0x73, 0x69, 0x67, 0x6e, 0x2e,
    0x63, 0x6f, 0x6d, 0x2f, 0x70, 0x63, 0x61, 0x33, 0x2d, 0x67, 0x35, 0x2e,
    0x63, 0x72, 0x6c, 0x30, 0x0e, 0x06, 0x03, 0x55, 0x1d, 0x0f, 0x01, 0x01,
    0xff, 0x04, 0x04, 0x03, 0x02, 0x01, 0x06, 0x30, 0x11, 0x06, 0x09, 0x60,
    0x86, 0x48, 0x01, 0x86, 0xf8, 0x42, 0x01, 0x01, 0x04, 0x04, 0x03, 0x02,
    0x01, 0x06, 0x30, 0x6d, 0x06, 0x08, 0x2b, 0x06, 0x01, 0x05, 0x05, 0x07,
    0x01, 0x0c, 0x04, 0x61, 0x30, 0x5f, 0xa1, 0x5d, 0xa0, 0x5b, 0x30, 0x59,
    0x30, 0x57, 0x30, 0x55, 0x16, 0x09, 0x69, 0x6d, 0x61, 0x67, 0x65, 0x2f,
    0x67, 0x69, 0x66, 0x30, 0x21, 0x30, 0x1f, 0x30, 0x07, 0x06, 0x05, 0x2b,
    0x0e, 0x03, 0x02, 0x1a, 0x04, 0x14, 0x8f, 0xe5, 0xd3, 0x1a, 0x86, 0xac,
    0x8d, 0x8e, 0x6b, 0xc3, 0xcf, 0x80, 0x6a, 0xd4, 0x48, 0x18, 0x2c, 0x7b,
    0x19, 0x2e, 0x30, 0x25, 0x16, 0x23, 0x68, 0x74, 0x74, 0x70, 0x3a, 0x2f,
    0x2f, 0x6c, 0x6f, 0x67, 0x6f, 0x2e, 0x76, 0x65, 0x72, 0x69, 0x73, 0x69,
    0x67, 0x6e, 0x2e, 0x63, 0x6f, 0x6d, 0x2f, 0x76, 0x73, 0x6c, 0x6f, 0x67,
    0x6f, 0x2e, 0x67, 0x69, 0x66, 0x30, 0x29, 0x06, 0x03, 0x55, 0x1d, 0x11,
    0x04, 0x22, 0x30, 0x20, 0xa4, 0x1e, 0x30, 0x1c, 0x31, 0x1a, 0x30, 0x18,
    0x06, 0x03, 0x55, 0x04, 0x03, 0x13, 0x11, 0x43, 0x6c, 0x61, 0x73, 0x73,
    0x33, 0x43, 0x41, 0x32, 0x30, 0x34, 0x38, 0x2d, 0x31, 0x2d, 0x34, 0x37,
    0x30, 0x3d, 0x06, 0x08, 0x2b, 0x06, 0x01, 0x05, 0x05, 0x07, 0x01, 0x01,
    0x04, 0x31, 0x30, 0x2f, 0x30, 0x2d, 0x06, 0x08, 0x2b, 0x06, 0x01, 0x05,
    0x05, 0x07, 0x30, 0x01, 0x86, 0x21, 0x68, 0x74, 0x74, 0x70, 0x3a, 0x2f,
    0x2f, 0x45, 0x56, 0x53, 0x65, 0x63, 0x75, 0x72, 0x65, 0x2d, 0x6f, 0x63,
    0x73, 0x70, 0x2e, 0x76, 0x65, 0x72, 0x69, 0x73, 0x69, 0x67, 0x6e, 0x2e,
    0x63, 0x6f, 0x6d, 0x30, 0x1f, 0x06, 0x03, 0x55, 0x1d, 0x23, 0x04, 0x18,
    0x30, 0x16, 0x80, 0x14, 0x7f, 0xd3, 0x65, 0xa7, 0xc2, 0xdd, 0xec, 0xbb,
    0xf0, 0x30, 0x09, 0xf3, 0x43, 0x39, 0xfa, 0x02, 0xaf, 0x33, 0x31, 0x33,
    0x30, 0x0d, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01,
    0x05, 0x05, 0x00, 0x03, 0x82, 0x01, 0x01, 0x00, 0x96, 0xa2, 0xfa, 0x7f,
    0xe6, 0x3d, 0xed, 0xd4, 0x2b, 0xce, 0xb7, 0x15, 0x3f, 0xc0, 0x72, 0x03,
    0x5f, 0x8b, 0xba, 0x16, 0x90, 0x25, 0xf7, 0xc2, 0x83, 0xd8, 0xc7, 0x75,
    0x34, 0x63, 0x68, 0x12, 0x53, 0x0c, 0x53, 0x89, 0x7b, 0xc9, 0x56, 0x09,
    0xa7, 0xc3, 0x36, 0x44, 0x4e, 0x0e, 0xd0, 0x62, 0x62, 0xb3, 0x86, 0xfa,
    0xe8, 0xa1, 0x9b, 0x34, 0x67, 0x8d, 0x53, 0x22, 0x17, 0x3e, 0xfd, 0xac,
    0xee, 0x67, 0x2e, 0x43, 0xe2, 0x5d, 0x7f, 0x33, 0x84, 0xf2, 0xa2, 0x70,
    0xc0, 0x6e, 0x82, 0x97, 0xc0, 0x34, 0xfd, 0x25, 0xc6, 0x23, 0x7f, 0xed,
    0xe6, 0xb0, 0xc5, 0x57, 0x43, 0x84, 0xb2, 0xde, 0x2d, 0xf1, 0xd0, 0xf6,
    0x48, 0x1f, 0x14, 0x71, 0x57, 0xb2, 0xac, 0x31, 0xe1, 0x97, 0x24, 0x23,
    0xc9, 0x13, 0x5d, 0x74, 0xe5, 0x46, 0xef, 0x09, 0x7c, 0x9e, 0xe1, 0x99,
    0x31, 0x0a, 0x08, 0x79, 0x1b, 0x8f, 0x71, 0x9f, 0x17, 0x66, 0xc8, 0x38,
    0xcf, 0xee, 0x8c, 0x97, 0xb6, 0x06, 0xb9, 0x73, 0x46, 0xe4, 0xd3, 0x94,
    0xc1, 0xe5, 0x60, 0xb5, 0x25, 0x75, 0x2d, 0xd9, 0x69, 0x31, 0xec, 0xcd,
    0x96, 0xc3, 0xa3, 0x76, 0xfd, 0xe8, 0x74, 0x44, 0xac, 0x12, 0xb9, 0x4d,
    0xbf, 0x51, 0xe8, 0xb9, 0xd4, 0x44, 0x4e, 0x27, 0xcb, 0xae, 0x20, 0xd1,
    0x7e, 0x2a, 0x7c, 0xb6, 0x63, 0x47, 0x9e, 0x76, 0xba, 0x97, 0xd0, 0x16,
    0xe7, 0x0b, 0x6c, 0x6d, 0xf7, 0x43, 0x6f, 0x33, 0x0b, 0x29, 0x30, 0x77,
    0xfa, 0x9d, 0xf9, 0xf5, 0x4e, 0xb8, 0x76, 0xb3, 0xcd, 0x18, 0xb4, 0xf9,
    0x20, 0xef, 0x3d, 0xdb, 0xe6, 0xca, 0xad, 0x9b, 0xd0, 0x4e, 0xd2, 0x87,
    0xa9, 0x0d, 0xa6, 0x44, 0x73, 0x50, 0xdd, 0x70, 0x5b, 0xed, 0xad, 0x7e,
    0x4a, 0xbc, 0x22, 0xd5, 0xa8, 0x26, 0xe4, 0xc2, 0x85, 0x20, 0x0d, 0xd9
};



/* A valid ocsp response */
static
uint8_t valid_ocsp_response_der[] = {
    0x30, 0x82, 0x01, 0x72, 0x0a, 0x01, 0x00, 0xa0, 0x82, 0x01, 0x6b, 0x30,
    0x82, 0x01, 0x67, 0x06, 0x09, 0x2b, 0x06, 0x01, 0x05, 0x05, 0x07, 0x30,
    0x01, 0x01, 0x04, 0x82, 0x01, 0x58, 0x30, 0x82, 0x01, 0x54, 0x30, 0x81,
    0xbe, 0xa1, 0x22, 0x30, 0x20, 0x31, 0x1e, 0x30, 0x1c, 0x06, 0x03, 0x55,
    0x04, 0x03, 0x13, 0x15, 0x63, 0x6f, 0x72, 0x65, 0x54, 0x4c, 0x53, 0x20,
    0x43, 0x41, 0x20, 0x43, 0x65, 0x72, 0x74, 0x20, 0x28, 0x52, 0x53, 0x41,
    0x29, 0x18, 0x0f, 0x32, 0x30, 0x31, 0x34, 0x30, 0x38, 0x32, 0x30, 0x32,
    0x31, 0x31, 0x33, 0x30, 0x37, 0x5a, 0x30, 0x62, 0x30, 0x60, 0x30, 0x3a,
    0x30, 0x09, 0x06, 0x05, 0x2b, 0x0e, 0x03, 0x02, 0x1a, 0x05, 0x00, 0x04,
    0x14, 0x20, 0xd4, 0x96, 0xb3, 0xfb, 0xd1, 0xb8, 0x84, 0x3a, 0x38, 0x14,
    0xdb, 0x33, 0xd1, 0x0d, 0xa8, 0xca, 0x96, 0xba, 0x13, 0x04, 0x14, 0xb2,
    0x23, 0x1b, 0x0f, 0x2c, 0x5a, 0xa2, 0x1d, 0xeb, 0x96, 0x34, 0xa7, 0x6f,
    0x9d, 0x97, 0x11, 0x81, 0x14, 0x61, 0xbb, 0x02, 0x01, 0x01, 0xa1, 0x11,
    0x18, 0x0f, 0x32, 0x30, 0x31, 0x34, 0x30, 0x38, 0x32, 0x30, 0x30, 0x30,
    0x30, 0x30, 0x30, 0x30, 0x5a, 0x18, 0x0f, 0x32, 0x30, 0x31, 0x34, 0x30,
    0x38, 0x32, 0x30, 0x32, 0x31, 0x31, 0x33, 0x30, 0x37, 0x5a, 0xa1, 0x23,
    0x30, 0x21, 0x30, 0x1f, 0x06, 0x09, 0x2b, 0x06, 0x01, 0x05, 0x05, 0x07,
    0x30, 0x01, 0x02, 0x04, 0x12, 0x04, 0x10, 0x4c, 0xc5, 0x63, 0xf2, 0x0a,
    0x84, 0x8c, 0x03, 0xa4, 0x0d, 0x97, 0xd1, 0xa2, 0xbb, 0x1e, 0xb2, 0x30,
    0x0d, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01, 0x05,
    0x05, 0x00, 0x03, 0x81, 0x81, 0x00, 0x1b, 0x21, 0xd7, 0x01, 0xde, 0xb8,
    0x58, 0x4b, 0x79, 0x6a, 0xa3, 0x8b, 0xa7, 0xe0, 0xbd, 0xa8, 0xda, 0x58,
    0x48, 0xbb, 0xa7, 0xcd, 0xf7, 0x91, 0x15, 0xb3, 0x38, 0x70, 0xd9, 0x43,
    0x25, 0x72, 0x0e, 0xc3, 0x3d, 0xf9, 0xc7, 0x30, 0x2d, 0xb4, 0x9f, 0x1c,
    0x4b, 0x62, 0x31, 0x48, 0xb4, 0x9f, 0x00, 0xbd, 0x57, 0xb6, 0xec, 0xda,
    0xf0, 0xa2, 0x42, 0x61, 0xfc, 0xef, 0x73, 0xc5, 0x55, 0xc1, 0xf6, 0x72,
    0x79, 0xcf, 0x55, 0x01, 0x09, 0xe4, 0xd2, 0xee, 0xbd, 0xa6, 0x08, 0xc6,
    0x39, 0x3a, 0x17, 0x76, 0x98, 0xaa, 0x61, 0x82, 0xb9, 0x41, 0xe1, 0xbb,
    0x4f, 0x67, 0x5e, 0x0b, 0x5e, 0xfa, 0x3c, 0x12, 0x15, 0xbe, 0x90, 0x8e,
    0x29, 0xe6, 0x5c, 0x9b, 0xfc, 0xaf, 0x40, 0xa4, 0x31, 0xd7, 0xa4, 0xc6,
    0x71, 0x22, 0x01, 0xfa, 0xb2, 0xcd, 0x6e, 0x1f, 0x26, 0xdb, 0xb1, 0xa3,
    0xec, 0x43
};


/* An ocsp response with a bad hash algorithm */
static
uint8_t bad_hash_ocsp_response_der[] = {
    0x30, 0x82, 0x01, 0x72, 0x0a, 0x01, 0x00, 0xa0, 0x82, 0x01, 0x6b, 0x30,
    0x82, 0x01, 0x67, 0x06, 0x09, 0x2b, 0x06, 0x01, 0x05, 0x05, 0x07, 0x30,
    0x01, 0x01, 0x04, 0x82, 0x01, 0x58, 0x30, 0x82, 0x01, 0x54, 0x30, 0x81,
    0xbe, 0xa1, 0x22, 0x30, 0x20, 0x31, 0x1e, 0x30, 0x1c, 0x06, 0x03, 0x55,
    0x04, 0x03, 0x13, 0x15, 0x63, 0x6f, 0x72, 0x65, 0x54, 0x4c, 0x53, 0x20,
    0x43, 0x41, 0x20, 0x43, 0x65, 0x72, 0x74, 0x20, 0x28, 0x52, 0x53, 0x41,
    0x29, 0x18, 0x0f, 0x32, 0x30, 0x31, 0x34, 0x30, 0x38, 0x32, 0x30, 0x32,
    0x31, 0x31, 0x33, 0x30, 0x37, 0x5a, 0x30, 0x62, 0x30, 0x60, 0x30, 0x3a,
    0x30, 0x09, 0x06, 0x05, 0x2b, 0x0e, 0x03, 0x02,/**/ 0x1B, /**/ 0x05, 0x00, 0x04,  // This line changed the OID for SHA1 to something else. (0x1B vs 0x1a)
    0x14, 0x20, 0xd4, 0x96, 0xb3, 0xfb, 0xd1, 0xb8, 0x84, 0x3a, 0x38, 0x14,
    0xdb, 0x33, 0xd1, 0x0d, 0xa8, 0xca, 0x96, 0xba, 0x13, 0x04, 0x14, 0xb2,
    0x23, 0x1b, 0x0f, 0x2c, 0x5a, 0xa2, 0x1d, 0xeb, 0x96, 0x34, 0xa7, 0x6f,
    0x9d, 0x97, 0x11, 0x81, 0x14, 0x61, 0xbb, 0x02, 0x01, /**/ 0x4B, /**/ 0xa1, 0x11, // This line change the serial number to match the cert (0x01 vs 0x4B)
    0x18, 0x0f, 0x32, 0x30, 0x31, 0x34, 0x30, 0x38, 0x32, 0x30, 0x30, 0x30,
    0x30, 0x30, 0x30, 0x30, 0x5a, 0x18, 0x0f, 0x32, 0x30, 0x31, 0x34, 0x30,
    0x38, 0x32, 0x30, 0x32, 0x31, 0x31, 0x33, 0x30, 0x37, 0x5a, 0xa1, 0x23,
    0x30, 0x21, 0x30, 0x1f, 0x06, 0x09, 0x2b, 0x06, 0x01, 0x05, 0x05, 0x07,
    0x30, 0x01, 0x02, 0x04, 0x12, 0x04, 0x10, 0x4c, 0xc5, 0x63, 0xf2, 0x0a,
    0x84, 0x8c, 0x03, 0xa4, 0x0d, 0x97, 0xd1, 0xa2, 0xbb, 0x1e, 0xb2, 0x30,
    0x0d, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01, 0x05,
    0x05, 0x00, 0x03, 0x81, 0x81, 0x00, 0x1b, 0x21, 0xd7, 0x01, 0xde, 0xb8,
    0x58, 0x4b, 0x79, 0x6a, 0xa3, 0x8b, 0xa7, 0xe0, 0xbd, 0xa8, 0xda, 0x58,
    0x48, 0xbb, 0xa7, 0xcd, 0xf7, 0x91, 0x15, 0xb3, 0x38, 0x70, 0xd9, 0x43,
    0x25, 0x72, 0x0e, 0xc3, 0x3d, 0xf9, 0xc7, 0x30, 0x2d, 0xb4, 0x9f, 0x1c,
    0x4b, 0x62, 0x31, 0x48, 0xb4, 0x9f, 0x00, 0xbd, 0x57, 0xb6, 0xec, 0xda,
    0xf0, 0xa2, 0x42, 0x61, 0xfc, 0xef, 0x73, 0xc5, 0x55, 0xc1, 0xf6, 0x72,
    0x79, 0xcf, 0x55, 0x01, 0x09, 0xe4, 0xd2, 0xee, 0xbd, 0xa6, 0x08, 0xc6,
    0x39, 0x3a, 0x17, 0x76, 0x98, 0xaa, 0x61, 0x82, 0xb9, 0x41, 0xe1, 0xbb,
    0x4f, 0x67, 0x5e, 0x0b, 0x5e, 0xfa, 0x3c, 0x12, 0x15, 0xbe, 0x90, 0x8e,
    0x29, 0xe6, 0x5c, 0x9b, 0xfc, 0xaf, 0x40, 0xa4, 0x31, 0xd7, 0xa4, 0xc6,
    0x71, 0x22, 0x01, 0xfa, 0xb2, 0xcd, 0x6e, 0x1f, 0x26, 0xdb, 0xb1, 0xa3,
    0xec, 0x43
};



/* A invalid ocsp response */
static uint8_t invalid_ocsp_response_der[] = {
    0x30, 0x81, 0x03,
    0x0a, 0x01, 0x01,
};

static void tests()
{
    SecCertificateRef certA=NULL, certD=NULL, certF=NULL, certCA_alpha=NULL, certCA_beta=NULL;
    CFDataRef proofD=NULL, proofA_1, proofA_2;
    SecCertificateRef www_digicert_com_cert=NULL, digicert_sha2_ev_server_ca=NULL;
    SecCertificateRef www_paypal_com_cert=NULL, www_paypal_com_issuer_cert=NULL;
    SecCertificateRef cfCert = NULL;
    CFMutableArrayRef certs=NULL;
    CFMutableArrayRef scts=NULL;
    CFMutableArrayRef anchors=NULL;
    CFDataRef valid_ocsp=NULL;
    CFDataRef invalid_ocsp=NULL;
    CFDataRef bad_hash_ocsp=NULL;

    CFArrayRef trustedLogs=NULL;
    CFURLRef trustedLogsURL=NULL;

    trustedLogsURL = CFBundleCopyResourceURL(CFBundleGetMainBundle(),
                                             CFSTR("si-82-sectrust-ct-logs"),
                                             CFSTR("plist"),
                                             NULL);
    isnt(trustedLogsURL, NULL, "trustedLogsURL");
    ok(CFURLResourceIsReachable(trustedLogsURL, NULL), "trustedLog plist is accessible");
    trustedLogs = (CFArrayRef) CFPropertyListReadFromFile(trustedLogsURL);


    isnt(certCA_alpha = SecCertificateCreateWithBytes(NULL, CA_alpha_cert_der, sizeof(CA_alpha_cert_der)), NULL, "create ca-alpha cert");
    isnt(certCA_beta = SecCertificateCreateWithBytes(NULL, CA_beta_cert_der, sizeof(CA_beta_cert_der)), NULL, "create ca-beta cert");
    isnt(anchors = CFArrayCreateMutable(kCFAllocatorDefault, 0, &kCFTypeArrayCallBacks), NULL, "create anchors array");
    CFArrayAppendValue(anchors, certCA_alpha);
    CFArrayAppendValue(anchors, certCA_beta);
    isnt(certA = SecCertificateCreateWithBytes(NULL, server_A_cert_der, sizeof(server_A_cert_der)), NULL, "create certD");
    isnt(certD = SecCertificateCreateWithBytes(NULL, serverD_cert_der, sizeof(serverD_cert_der)), NULL, "create certD");
    isnt(certF = SecCertificateCreateWithBytes(NULL, serverF_cert_der, sizeof(serverF_cert_der)), NULL, "create certF");
    isnt(proofD = CFDataCreate(kCFAllocatorDefault, serverD_cert_proof, sizeof(serverD_cert_proof)), NULL, "creat proofD");
    isnt(proofA_1 = CFDataCreate(kCFAllocatorDefault, server_A_proof_Alfa_3, sizeof(server_A_proof_Alfa_3)), NULL, "creat proofA_1");
    isnt(proofA_2 = CFDataCreate(kCFAllocatorDefault, server_A_proof_Bravo_3, sizeof(server_A_proof_Bravo_3)), NULL, "creat proofA_2");
    isnt(www_digicert_com_cert = SecCertificateCreateWithBytes(NULL, www_digicert_com_cert_der, sizeof(www_digicert_com_cert_der)), NULL, "create www.digicert.com cert");
    isnt(digicert_sha2_ev_server_ca = SecCertificateCreateWithBytes(NULL, digicert_sha2_ev_server_ca_der, sizeof(digicert_sha2_ev_server_ca_der)), NULL, "create digicert.com subCA cert");
    isnt(www_paypal_com_cert = SecCertificateCreateWithBytes(NULL, www_paypal_com_cert_der, sizeof(www_paypal_com_cert_der)), NULL, "create www.paypal.com cert");
    isnt(www_paypal_com_issuer_cert = SecCertificateCreateWithBytes(NULL, www_paypal_com_issuer_cert_der, sizeof(www_paypal_com_issuer_cert_der)), NULL, "create www.paypal.com issuer cert");
    isnt(valid_ocsp = CFDataCreate(kCFAllocatorDefault, valid_ocsp_response_der, sizeof(valid_ocsp_response_der)), NULL, "create valid_ocsp");
    isnt(invalid_ocsp = CFDataCreate(kCFAllocatorDefault, invalid_ocsp_response_der, sizeof(invalid_ocsp_response_der)), NULL, "create invalid_ocsp");
    isnt(bad_hash_ocsp = CFDataCreate(kCFAllocatorDefault, bad_hash_ocsp_response_der, sizeof(bad_hash_ocsp_response_der)), NULL, "create bad_hash_ocsp");

    /* Case 1: coreos-ct-test embedded SCT - only 1 SCT - so not CT qualified */
    isnt(certs = CFArrayCreateMutable(kCFAllocatorDefault, 0, &kCFTypeArrayCallBacks), NULL, "create cert array");
    CFArrayAppendValue(certs, certF);
    test_ct_trust(certs, NULL, NULL, anchors, trustedLogs, CFSTR("coreos-ct-test.apple.com"), false, "coreos-ct-test 1");
    CFReleaseNull(certs);

    /* Case 2: coreos-ct-test standalone SCT - only 1 SCT - so not CT qualified  */
    isnt(certs = CFArrayCreateMutable(kCFAllocatorDefault, 0, &kCFTypeArrayCallBacks), NULL, "create cert array");
    CFArrayAppendValue(certs, certD);
    isnt(scts = CFArrayCreateMutable(kCFAllocatorDefault, 0, &kCFTypeArrayCallBacks), NULL, "create SCT array");
    CFArrayAppendValue(scts, proofD);
    test_ct_trust(certs, scts, NULL, anchors, trustedLogs, CFSTR("coreos-ct-test.apple.com"), false, "coreos-ct-test 2");
    CFReleaseNull(certs);
    CFReleaseNull(scts);

    /* case 3: digicert : 2 embedded SCTs, but both from Google, so not CT qualified */
    isnt(certs = CFArrayCreateMutable(kCFAllocatorDefault, 0, &kCFTypeArrayCallBacks), NULL, "create cert array");
    CFArrayAppendValue(certs, www_digicert_com_cert);
    CFArrayAppendValue(certs, digicert_sha2_ev_server_ca);
    test_ct_trust(certs, NULL, NULL, NULL, NULL, CFSTR("www.digicert.com"), false, "digicert");
    CFReleaseNull(certs);

    /* case 4: paypal.com cert - not CT, but EV */
    isnt(certs = CFArrayCreateMutable(kCFAllocatorDefault, 0, &kCFTypeArrayCallBacks), NULL, "create cert array");
    CFArrayAppendValue(certs, www_paypal_com_cert);
    CFArrayAppendValue(certs, www_paypal_com_issuer_cert);
    test_ct_trust(certs, NULL, NULL, NULL, trustedLogs, CFSTR("www.paypal.com"), false, "paypal");
    CFReleaseNull(certs);

    /* Case 5: coreos-ct-test standalone SCT -  2 SCTs - CT qualified  */
    isnt(certs = CFArrayCreateMutable(kCFAllocatorDefault, 0, &kCFTypeArrayCallBacks), NULL, "create cert array");
    CFArrayAppendValue(certs, certA);
    isnt(scts = CFArrayCreateMutable(kCFAllocatorDefault, 0, &kCFTypeArrayCallBacks), NULL, "create SCT array");
    CFArrayAppendValue(scts, proofA_1);
    CFArrayAppendValue(scts, proofA_2);
    test_ct_trust(certs, scts, NULL, anchors, trustedLogs, CFSTR("coreos-ct-test.apple.com"), true, "coreos-ct-test 3");
    CFReleaseNull(certs);
    CFReleaseNull(scts);


    /* Case 6: Test with an invalid OCSP response */
    isnt(certs = CFArrayCreateMutable(kCFAllocatorDefault, 0, &kCFTypeArrayCallBacks), NULL, "create cert array");
    CFArrayAppendValue(certs, certA);
    isnt(scts = CFArrayCreateMutable(kCFAllocatorDefault, 0, &kCFTypeArrayCallBacks), NULL, "create SCT array");
    CFArrayAppendValue(scts, proofA_1);
    test_ct_trust(certs, scts, invalid_ocsp, anchors, trustedLogs, CFSTR("coreos-ct-test.apple.com"), false, "coreos-ct-test 3");
    CFReleaseNull(certs);
    CFReleaseNull(scts);

    /* Case 7: Test with a valid OCSP response */
    isnt(certs = CFArrayCreateMutable(kCFAllocatorDefault, 0, &kCFTypeArrayCallBacks), NULL, "create cert array");
    CFArrayAppendValue(certs, certA);
    isnt(scts = CFArrayCreateMutable(kCFAllocatorDefault, 0, &kCFTypeArrayCallBacks), NULL, "create SCT array");
    CFArrayAppendValue(scts, proofA_1);
    test_ct_trust(certs, scts, valid_ocsp, anchors, trustedLogs, CFSTR("coreos-ct-test.apple.com"), false, "coreos-ct-test 3");
    CFReleaseNull(certs);
    CFReleaseNull(scts);

    /* Case 8: Test with a bad hash OCSP response */
    isnt(certs = CFArrayCreateMutable(kCFAllocatorDefault, 0, &kCFTypeArrayCallBacks), NULL, "create cert array");
    CFArrayAppendValue(certs, certA);
    isnt(scts = CFArrayCreateMutable(kCFAllocatorDefault, 0, &kCFTypeArrayCallBacks), NULL, "create SCT array");
    CFArrayAppendValue(scts, proofA_1);
    test_ct_trust(certs, scts, bad_hash_ocsp, anchors, trustedLogs, CFSTR("coreos-ct-test.apple.com"), false, "coreos-ct-test 3");
    CFReleaseNull(certs);
    CFReleaseNull(scts);


#define TEST_CASE(x) \
 do { \
    isnt(certs = CFArrayCreateMutable(kCFAllocatorDefault, 0, &kCFTypeArrayCallBacks), NULL, "create cert array for " #x); \
    isnt(cfCert = SecCertificateCreateWithBytes(NULL, x, sizeof(x)), NULL, "create cfCert from " #x); \
    CFArrayAppendValue(certs, cfCert); \
    test_ct_trust(certs, NULL, NULL, anchors, trustedLogs, CFSTR("coreos-ct-test.apple.com"), true, #x); \
    CFReleaseNull(certs);  \
    CFReleaseNull(cfCert);  \
 } while (0)

/*
 server_1601_cert_der
 server_1603_cert_der
 server_1604_cert_der
 server_1701_cert_der
 server_1704_cert_der
 server_1705_cert_der
 server_1801_cert_der
 server_1804_cert_der
 server_1805_cert_der
 server_2001_cert_der
*/

    TEST_CASE(server_1601_cert_der);
    TEST_CASE(server_1603_cert_der);
    TEST_CASE(server_1604_cert_der);
    TEST_CASE(server_1701_cert_der);
    TEST_CASE(server_1704_cert_der);
    TEST_CASE(server_1705_cert_der);
    TEST_CASE(server_1801_cert_der);
    TEST_CASE(server_1804_cert_der);
    TEST_CASE(server_1805_cert_der);
    TEST_CASE(server_2001_cert_der);

    CFReleaseSafe(certCA_alpha);
    CFReleaseSafe(certCA_beta);
    CFReleaseSafe(anchors);
    CFReleaseSafe(certA);
    CFReleaseSafe(certD);
    CFReleaseSafe(certF);
    CFReleaseSafe(proofD);
    CFReleaseSafe(proofA_1);
    CFReleaseSafe(proofA_2);
    CFReleaseSafe(www_digicert_com_cert);
    CFReleaseSafe(digicert_sha2_ev_server_ca);
    CFReleaseSafe(www_paypal_com_cert);
    CFReleaseSafe(www_paypal_com_issuer_cert);
    CFReleaseSafe(trustedLogsURL);
    CFReleaseSafe(trustedLogs);
    CFReleaseSafe(valid_ocsp);
    CFReleaseSafe(invalid_ocsp);
    CFReleaseSafe(bad_hash_ocsp);

}


int si_82_sectrust_ct(int argc, char *const *argv)
{
	plan_tests(287);

	tests();

	return 0;
}
